#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **dataclass decorators** (i.e., low-level decorators decorating
pure-Python types decorated by the :pep:`557`-compliant
:obj:`dataclasses.dataclass` decorator).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeCallHintPep557FieldViolation
from beartype._conf.confmain import BeartypeConf
from beartype._data.typing.datatypingport import (
    DictStrToHint,
    Hint,
)
from beartype._data.hint.sign.datahintsignset import (
    HINT_SIGNS_DATACLASS_NONFIELDS)
from beartype._data.kind.datakindiota import SENTINEL
from beartype._util.cls.pep.clspep557 import (
    die_unless_type_pep557_dataclass,
    is_pep557_dataclass_frozen,
)
from beartype._util.cls.utilclsset import set_type_attr
from beartype._util.hint.pep.proposal.pep649 import (
    get_pep649_hintable_annotations)
from beartype._util.hint.pep.utilpepsign import get_hint_pep_sign_or_none
from beartype._util.utilobject import get_object_type_name

# ....................{ DECORATORS                         }....................
#FIXME: As a mandatory prerequisite *BEFORE* integrating this into the @beartype
#codebase, we first need to:
#* Generalize both is_bearable() and die_if_unbearable() to support quoted
#  relative forward references. As always, the algorithm should iteratively
#  search up the callstack for the first stack frame residing *OUTSIDE*
#  @beartype. Actually... that doesn't suffice. Third-party frameworks
#  leveraging @beartype could themselves be consuming other third-party type
#  hints originating from users. Deciding where exactly those type hints were
#  originally defined is *PROBABLY* infeasible in the general case. So, we
#  really do need to iteratively search up the entire call stack before raising
#  an exception. It's fine. Just do it. The alternative is broken badness.
#FIXME: Unit test against all possible dataclass edge cases, including:
#* Quoted relative forward references (e.g., "list['MuhUndefinedType']").
#* "typing.Self". We're *NOT* passing "cls_stack" to either the is_bearable() or
#  die_if_unbearable() functions, because those functions currently fail to
#  accept an optional "cls_stack" parameter. We should probably generalize both
#  of those functions to accept that parameter, huh? *sigh*
#* Dataclass subclasses. Does each dataclass subclass in a hierarchy have its
#  own unique "__annotations__" dunder dictionary *OR* does each such subclass
#  composite the "__annotations__" of both itself and its superclasses? Probably
#  the former, huh? Yikes. This implies that our trivial attempt to directly use
#  "__annotations__" fails to suffice. We'll actually have to iteratively crawl
#  up "datacls.__mro__" and composite the full "__annotations__" from the
#  "__annotations__" of all dataclass superclasses.
#
#  Note, however, that there exists a critical optimization here: we
#  *ABSOLUTELY* need to stop iterating up "datacls.__mro__" when we visit the
#  first superclass that is *NOT* also a dataclass. Is that even possible? No
#  idea. If it is, we halt iteration at that first non-dataclass. This is
#  essential, as the class attributes of that first non-dataclass (and all
#  superclasses of that non-dataclass) *CANNOT* by definition by fields.
#
#  *WAIT.* Halting iteration doesn't work, because "datacls.__mro__" doesn't
#  exactly correspond to superclass relations. Consider diamond inheritance, for
#  example. Ergo, we'll instead have to inefficiently *IGNORE* all superclasses
#  in "datacls.__mro__" that are *NOT* themselves dataclasses. Fine. No worries.
#  We can certainly do that. Nonetheless, we sigh. *sigh*
#
#  Oh -- and note that we'll need to iteratively resolve PEP 563-postponed
#  stringified type hints against each such superclass "__annotations__" as
#  well. Jeez. This sure got ugly fast, huh? So much sighing! *sigh sigh*
#* PEP 563, subject to the constraints detailed above.
def beartype_pep557_dataclass(
    # Mandatory parameters.
    #
    # Note that dataclasses do *NOT* have a more specific superclass than merely
    # the root "type" superclass of *ALL* types, sadly:
    #     >>> from dataclasses import dataclass
    #     >>> @dataclass
    #     ... class MuhDataclass(object): muh_int: int
    #     >>> MuhDataclass.__mro__
    #     (<class '__main__.MuhDataclass'>, <class 'object'>)  # <--- yikes
    #
    # From a typing perspective, "type" is the best that can be done. Yikes!
    datacls: type,
    conf: BeartypeConf,

    # Optional parameters.
    exception_prefix: str = '',
) -> None:
    '''
    Decorate the passed **dataclass** (i.e., pure-Python class decorated by the
    :pep:`557`-compliant :obj:`dataclasses.dataclass` decorator) with
    dynamically generated type-checking of all **dataclass fields** (i.e., class
    attributes annotated by *any* type hints other than :pep:`526`-compliant
    ``dataclasses.ClassVar[...]`` or :pep:`557`-compliant
    ``dataclasses.InitVar[...]`` type hints) on both **dataclass object
    initialization** (i.e., at ``__init__()`` time) *and* **dataclass field
    assignment** (i.e., when each field is subsequently assigned to by an
    assignment statement).

    This decorator *only* type-checks **dataclass fields.** By :pep:`557`, a
    "dataclass field" is *any* class attribute of this dataclass annotated by
    *any* type hint other than either:

    * A :pep:`526`-compliant ``dataclasses.ClassVar[...]`` type hint.
    * A :pep:`557`-compliant ``dataclasses.InitVar[...]`` type hint.

    Unlike most :mod:`beartype` decorators, this decorator safely monkey-patches
    this dataclass in-place. (Equivalently, this decorator safely monkey-patches
    type-checking into this same dataclass *without* creating or returning a new
    dataclass.) Specifically, this decorator monkey-patches the
    ``__setattr__()`` dunder method of this dataclass. If this dataclass does
    *not* directly define ``__setattr__()``, this decorator adds a new
    ``__setattr__()`` to this dataclass; else, this decorator wraps the existing
    ``__setattr__()`` already directly defined on this dataclass with a new
    ``__setattr__()`` internally deferring to that existing ``__setattr__()``.
    In either case, this new ``__setattr__()`` type-checks that each dataclass
    field satisfies the type hint annotating that field on both:

    * **Dataclass object initialization** (i.e., at early ``__init__()`` time).
    * **Dataclass field assignment** (i.e., when each field is subsequently
      assigned to by an assignment statement).

    Parameters
    ----------
    datacls : BeartypeableT
        Dataclass to be decorated by :func:`beartype.beartype`.
    conf : BeartypeConf
        Beartype configuration configuring :func:`beartype.beartype` uniquely
        specific to this dataclass.
    exception_prefix : str, default: ''
        Human-readable substring prefixing raised exceptions messages. Defaults
        to the empty string.

    Returns
    -------
    BeartypeableT
        This same dataclass monkey-patched in-place with type-checking.
    '''
    assert isinstance(conf, BeartypeConf), (
        f'{repr(conf)} not beartype configuration.')

    # ..................{ PREAMBLE                           }..................
    # If this dataclass is *NOT* actually a dataclass, raise an exception.
    die_unless_type_pep557_dataclass(
        cls=datacls, exception_prefix=exception_prefix)
    # Else, this dataclass is actually a dataclass.

    # ..................{ IMPORTS                            }..................
    # Defer heavyweight imports prohibited at global scope.
    from beartype.door import (
        die_if_unbearable,
        is_bearable,
    )

    # ..................{ LOCALS                             }..................
    # *HORRIBLE HACK*. For unknown reasons, the super() function called below
    # requires the "__class__" attribute to be defined as a cell (i.e., closure)
    # variable. If this is *NOT* the case, then that call raises the unreadable
    # low-level exception:
    #     RuntimeError: super(): __class__ cell not found
    __class__ = datacls

    # __setattr__() dunder method directly defined on this dataclass if any *OR*
    # "None" (i.e., if this dataclass does *NOT* directly define this method).
    datacls_setattr = datacls.__dict__.get('__setattr__')

    # ..................{ SANIFICATION                       }..................
    # Sanify (i.e., sanitize) this dictionary of type hints.

    #FIXME: Copy this dictionary via dict.copy() for safety. Directly modifying
    #"__annotations__" dunder dictionaries is probably unsafe in Python >= 3.14.
    #Since this is becoming a common operation, perhaps simply add a new
    #optional "is_copy: bool = False" parameter to this get_object_annotations()
    #getter. If "is_copy" is true, then that getter performs the copy for us.

    # Unsanified (i.e., original) dictionary mapping from the name of each
    # possible field of this dataclass to the possibly insane type hint
    # annotating that field *AFTER* resolving all PEP 563-postponed type hints.
    attr_name_to_hint_insane = get_pep649_hintable_annotations(datacls)

    # Sanified (i.e., sanitized) dictionary mapping from the name of each
    # guaranteable field of this dataclass to the ostensibly sane type hint
    # annotating that field, initialized to the empty dictionary.
    field_name_to_hint: DictStrToHint = {}

    # dict.get() method bound to this dictionary as a negligible optimization.
    field_name_to_hint_get = field_name_to_hint.get

    #FIXME: Note that an edge case could arise here under Python >= 3.12 due to
    #the intersection of PEP 563 and 695:
    #    from __future__ import annotations  # <-- PEP 563
    #    from dataclasses import dataclass
    #
    #    type ohnoes[T] = T | int  # <-- PEP 695
    #
    #    @dataclass
    #    class Ugh(object):
    #        guh: ohnoes[str]
    #
    #To efficiently resolve this, we *PROBABLY* want to generalize our existing
    #beartype.peps.resolve_pep563() resolver to additionally support types in
    #addition to its existing support for classes. Naturally, this gets ugly
    #fast. For example:
    #* The existing resolve_pep563() function accepts a "func" parameter.
    #  Consider deprecating this parameter and instead requesting that callers
    #  pass only a generic "obj" parameter.
    #* Generalize this function to accept an "obj" parameter resembling:
    #      obj: Annotationsable

    # For the name and unsanified hint of each class attribute of this
    # dataclass...
    for field_name, field_hint in attr_name_to_hint_insane.items():
        # Sign uniquely identifying this unsanified hint.
        field_hint_sign = get_hint_pep_sign_or_none(field_hint)

        # If this sign signifies this class attribute to *NOT* be a dataclass
        # field, remove this attribute from consideration by ignoring this
        # attribute rather than adding this attribute back to this dictionary.
        #
        # Note that attempting to identify unsanified hints is often a bad idea.
        # Only sanified hints are safely identifiable, usually. This might be
        # the one and only edge case where identifying an unsanified hint is not
        # only reasonable but desirable. Why? PEP 557, which explicitly states
        # that both PEP 526-compliant "type.ClassVar[...]" *AND* PEP
        # 557-compliant "dataclasses.InitVar[...]" hints are only valid as root
        # hints directly annotating class variables of dataclasses. Why? Because
        # the PEP 557-compliant @dataclasses.dataclass decorator itself
        # explicitly detects these root hints with a crude detection scheme that
        # only works because these hints are required to be root. Since PEP
        # 563-postponed stringified type hints are guaranteed to have already
        # been resolved above, these hints are guaranteed to be both
        # non-stringified and root hints. W00t!
        if field_hint_sign in HINT_SIGNS_DATACLASS_NONFIELDS:
            continue
        # Else, this sign signifies this class attribute to actually be a field.

        #FIXME: Insufficient. We also need to immediately sanify *ALL* of these
        #hints right here *OUTSIDE* of the closure defined below. Yet again,
        #issues arise. Why? Because the sanify_hint_root_func() function is
        #inappropriate here. Instead:
        #* Define a new sanify_hint_root_type() getter. This could prove
        #  non-trivial. sanify_hint_root_func() accepts a "decor_meta"
        #  parameter, which currently only applies to decorated *CALLABLES*
        #  rather than *TYPES*. We probably want to generalize "decor_meta" to
        #  support both... maybe? Maybe not? To do this properly, we probably
        #  first want to:
        #  * Create a new "decor_meta" type hierarchy resembling:
        #        class BeartypeDecorMetaABC(metaclass=ABCMeta): ...
        #        class BeartypeDecorMetaFunc(BeartypeDecorMetaABC): ...
        #        class BeartypeDecorMetaType(BeartypeDecorMetaABC): ...
        #  * Refactor references to "BeartypeDecorMeta" to either
        #    "BeartypeDecorMetaABC" *OR* ""BeartypeDecorMetaFunc" depending on
        #    context. Most probably require the latter. Any that don't should
        #    simply reference "BeartypeDecorMetaABC" for generality.
        #  * Remove all references to "BeartypeDecorMeta".
        #FIXME: Consider:
        #* If sanifying this hint so reduced this hint to "Any", remove this
        #  hint from this dictionary entirely. Doing so speeds up closure logic
        #  below, which is critical.
        #* Actually... this could be a problematic approach. Why?
        #  "hint_or_sane", of course. is_bearable() and die_if_unbearable() only
        #  accept actual type hints. But "hint_or_sane" could be a
        #  @beartype-specific type hint dataclass! So... that doesn't quite
        #  work. I suppose what we could do is an optimization resembling:
        #  * If sanifying this hint produced a different type hint than the
        #    original type hint *AND* this new type hint is *NOT* simply a
        #    "HintSane" object, replace this old hint with this new hint
        #    in the "field_name_to_hint" dictionary.
        #  * Else, preserve this existing hint in this dictionary as is. If a
        #    "HintSane" object was produced, we'll just have to throw
        #    that away for the moment. Alternately, we could *TRY* to generalize
        #    is_bearable() and die_if_unbearable() to accept these objects.
        #    But... probably not worth it for the moment. It is what it is.
        #* Actually... we can do something even better! There's no particular
        #  reason we have to call the public-facing is_bearable() and
        #  die_if_unbearable() functions. Instead:
        #  * Define new private-facing variants of those functions transparently
        #    accepting a "hint: HintSane" parameter. Call them:
        #    * is_hint_sane_bearable().
        #    * die_if_hint_sane_unbearable().
        #    In theory, this shouldn't be *TOO* hard.
        #  * Call these private- rather than public-facing variants below.
        #    Voila! Problem transparently resolved.

        # Add this field back to this sanified dictionary.
        field_name_to_hint[field_name] = field_hint

    # ..................{ CLOSURES                           }..................
    def check_pep557_dataclass_field(
        self, attr_name: str, attr_value: object) -> None:
        # Type hint annotating this dataclass attribute if this attribute is
        # annotated and thus (probably) a dataclass field *OR* "None" otherwise
        # (i.e., if this attribute is unannotated).
        #
        # Note that:
        # * There exists a (mostly) one-to-one correlation between fields and
        #   type hints. PEP 557 literally defines a dataclass field as an
        #   annotated dataclass attribute:
        #      A field is defined as any variable identified in __annotations__.
        #      That is, a variable that has a type annotation.
        # * There exist alternate means of introspecting dataclass fields (e.g.,
        #   the public dataclasses.fields() global function). Without exception,
        #   these alternates are all less efficient *AND* more cumbersome than
        #   simply directly introspecting dataclass field type hints. Moreover,
        #   these alternates are unlikely to play nicely with unquoted forward
        #   references under Python >= 3.14.
        attr_hint: Hint = field_name_to_hint_get(attr_name, SENTINEL)  # type: ignore[arg-type]

        # If this dataclass attribute is annotated and thus a field...
        if attr_hint is not SENTINEL:
            # If the new value of this field violates this hint...
            #
            # Note that this is a non-negligible optimization. Technically, this
            # preliminary test is superfluous: only the call to the
            # die_if_unbearable() raiser below is required. Pragmatically, this
            # preliminary test avoids various needlessly expensive operations in
            # the common case that this value satisfies this hint.
            if not is_bearable(obj=attr_value, hint=attr_hint, conf=conf):  # pyright: ignore
                #FIXME: *UGLY LOGIC.* Sure. Technically, this works. But we
                #repeat the *EXACT* same logic in our currently unused
                #_die_if_arg_default_unbearable() validator, which we will
                #almost certainly re-enable at some point. Instead:
                #* Just add a new optional "exception_cls" parameter to the
                #  die_if_unbearable() validator called below. If necessary, the
                #  initial implementation of this parameter could just do what
                #  we currently do here. Not great, but at least that logic
                #  would be centralized away from prying eyes in the same API.

                # Modifiable keyword dictionary encapsulating this beartype
                # configuration.
                conf_kwargs = conf.kwargs.copy()

                #FIXME: This should probably be configurable as well. For now,
                #this is fine. We shrug noncommittally. We shrug, everyone!
                # Set the type of violation exception raised by the subsequent
                # call to the die_if_unbearable() function to the expected type.
                conf_kwargs['violation_door_type'] = (
                    BeartypeCallHintPep557FieldViolation)

                # New beartype configuration initialized by this dictionary.
                conf_new = BeartypeConf(**conf_kwargs)

                # Machine-readable representation of this dataclass instance.
                self_repr: str = ''

                # Attempt to introspect this representation of this instance.
                # There exist two common cases here:
                # * This instance has already been fully initialized (i.e., the
                #   __init__() dunder method has already successfully returned),
                #   implying that all dataclass fields have already been set to
                #   valid values on this instance. In this case, this
                #   "repr(self)" call *SHOULD* succeed -- unless this dataclass
                #   subclass has erroneously redefined the __repr__() dunder
                #   method in a fragile manner raising unexpected exceptions.
                # * This instance has *NOT* yet been fully initialized (i.e.,
                #   the __init__() dunder method has *NOT* yet successfully
                #   returned), implying that one or more dataclass fields have
                #   *NOT* yet been set to valid values on this instance. In this
                #   case, this "repr(self)" call *SHOULD* fail with an
                #   "AttributeError" resembling:
                #       AttributeError: '{class_name}' object has no attribute '{attr_name}'
                try:
                    self_repr = repr(self)
                # If introspecting this representation fails for any reason
                # whatsoever, fallback to just the fully-qualified name of this
                # dataclass subclass, which should *ALWAYS* be introspectable.
                except Exception:
                    self_repr = repr(get_object_type_name(datacls))

                # Human-readable substring prefixing the exception raised below.
                #
                # Note that the die_if_unbearable() raiser implicitly suffixes
                # this prefix by the substring "value". On the one hand, it
                # probably shouldn't be doing that. On the other hand, it
                # currently is doing that. On the gripping hand, we're too tired
                # to do anything about it doing that. This is why bugs exist.
                exception_prefix = (
                    f'Dataclass {self_repr} '
                    f'attribute {repr(attr_name)} new '
                )

                # Raise this type of violation exception.
                die_if_unbearable(
                    obj=attr_value,
                    hint=attr_hint,
                    conf=conf_new,
                    exception_prefix=exception_prefix,
                )
            # Else, the new value of this field satisfies this hint. In this
            # case, silently reduce to a noop.
        # Else, this dataclass attribute is unannotated and thus *NOT* a field.
        # In this case, this attribute is ignorable.

        # If this dataclass does *NOT* directly override the superclass
        # __setattr__() dunder method with a non-default dataclass-specific
        # __setattr__() dunder method, fallback to the former. The superclass
        # __setattr__() dunder method is guaranteed to be defined on at least
        # one superclass of this dataclass. Why? Because the root superclass
        # type.__setattr__() dunder method is guaranteed to exist on all types.
        #
        # Note that:
        # * This is the common case and thus tested first.
        # * Unlike the below case, this method method is accessed via the
        #   super() builtin and is thus a true method bound to this dataclass.
        #   Ergo, the "self" parameter must *NOT* be explicitly passed.
        if datacls_setattr is None:
            super().__setattr__(attr_name, attr_value)  # type: ignore[misc]
        # Else, this dataclass directly defines a non-default dataclass-specific
        # implementation of this method overriding the superclass __setattr__()
        # dunder method. In this case, defer to this override.
        #
        # Note that, unlike the above case, this method was accessed via the
        # "__dict__" dunder dictionary and is thus an unbound function *NOT*
        # bound to this dataclass. Ergo, the "self" parameter *MUST* be
        # explicitly passed.
        else:
            datacls_setattr(self, attr_name, attr_value)

    # ..................{ DECORATORS                         }..................
    # setattr()-like callable to be called to set this attribute on this
    # dataclass, defined as either...
    setattr_func = (
        # If this dataclass is frozen, the standard setattr() builtin does *NOT*
        # suffice. Why? Because frozen dataclasses guarantee immutability by
        # overriding the __setattr__() dunder method (implicitly called by the
        # setattr() builtin) to unconditionally raise an exception. While
        # understandable, this behaviour prevents the set_type_attr() function
        # called below from monkey-patching type-checking into this dataclass;
        # attempting to do so would ironically invoke that same __setattr__()
        # dunder method, which would then raises an exception. This behaviour
        # can be circumvented by passing the type.__setattr__() dunder method as
        # this parameter, which then applies the desired monkey-patch *WITHOUT*
        # raising an exception. In short, stupid kludges is always the answer.
        type.__setattr__
        if is_pep557_dataclass_frozen(
            datacls=datacls, exception_prefix=exception_prefix) else
        # Else, this dataclass is *NOT* frozen. In this case, the standard
        # setattr() builtin, which internally defers to the __setattr__() dunder
        # method guaranteed to be defined by all dataclasses (due to the
        # existence of the type.__setattr__() dunder method).
        setattr
    )

    # Safely replace this undecorated __setattr__() implementation with this
    # decorated __setattr__() implementation.
    set_type_attr(
        cls=datacls,
        attr_name='__setattr__',
        attr_value=check_pep557_dataclass_field,
        setattr_func=setattr_func,  # pyright: ignore
    )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **class tester** unit tests.

This submodule unit tests the public API of the private
:mod:`beartype._util.cls.utilclstest` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS ~ tester                     }....................
def test_is_type_or_types() -> None:
    '''
    Test the :func:`beartype._util.cls.utilclstest.is_type_or_types` tester.
    '''

    # Defer test-specific imports.
    from beartype._util.cls.utilclstest import is_type_or_types

    # Assert this tester accepts an arbitrary type.
    assert is_type_or_types(str) is True

    # Assert this tester accepts an arbitrary non-empty tuple of types.
    assert is_type_or_types((bool, int)) is True

    # Assert this tester rejects an arbitrary object that is neither a type
    # nor a non-empty tuple of types.
    assert is_type_or_types(
        'To drink their odours, and their mighty swinging') is False

    # Assert this tester rejects the empty tuple.
    assert is_type_or_types(()) is False

    # Assert this tester rejects an arbitrary non-empty tuple containing one or
    # more non-types.
    assert is_type_or_types(
        (dict, 'To hear—an old and solemn harmony;', float)) is False


def test_is_type_subclass() -> None:
    '''
    Test the :func:`beartype._util.cls.utilclstest.is_type_subclass` tester.
    '''

    # Defer test-specific imports.
    from beartype._util.cls.utilclstest import is_type_subclass
    from beartype_test.a00_unit.data.data_type import Class, Subclass

    # Assert this tester accepts objects that are subclasses of superclasses.
    assert is_type_subclass(Subclass, Class) is True
    assert is_type_subclass(Class, object) is True
    assert is_type_subclass(Class, (Class, str)) is True

    # Assert this tester rejects objects that are superclasses of subclasses.
    assert is_type_subclass(object, Class) is False
    assert is_type_subclass(Class, Subclass) is False
    assert is_type_subclass(Class, (Subclass, str)) is False

    # Assert this tester rejects objects that are unrelated classes.
    assert is_type_subclass(str, bool) is False

    # Assert this tester rejects objects that are non-classes *WITHOUT* raising
    # an exception.
    assert is_type_subclass(
        "Thou many-colour'd, many-voiced vale,", str) is False

# ....................{ TESTS ~ tester : builtin           }....................
def test_is_type_builtin() -> None:
    '''
    Test the :func:`beartype._util.cls.utilclstest.is_type_builtin` tester.
    '''

    # Defer test-specific imports.
    from beartype._util.cls.utilclstest import is_type_builtin
    from beartype_test.a00_unit.data.data_type import (
        TYPES_BUILTIN,
        TYPES_NONBUILTIN,
    )

    # Assert this tester accepts all builtin types.
    for type_builtin in TYPES_BUILTIN:
        assert is_type_builtin(type_builtin) is True

    # Assert this tester rejects non-builtin types.
    for type_nonbuiltin in TYPES_NONBUILTIN:
        assert is_type_builtin(type_nonbuiltin) is False


def test_is_type_builtin_or_fake() -> None:
    '''
    Test the :func:`beartype._util.cls.utilclstest.is_type_builtin_or_fake`
    tester.
    '''

    # Defer test-specific imports.
    from beartype._util.cls.utilclstest import is_type_builtin_or_fake
    from beartype_test.a00_unit.data.data_type import (
        TYPES_BUILTIN,
        TYPES_BUILTIN_FAKE,
        Class,
    )

    # Assert this tester accepts all non-fake builtin types.
    for type_builtin in TYPES_BUILTIN:
        assert is_type_builtin_or_fake(type_builtin) is True

    # Assert this tester accepts all fake builtin types, too.
    for type_builtin_fake in TYPES_BUILTIN_FAKE:
        assert is_type_builtin_or_fake(type_builtin_fake) is True

    # Assert this tester rejects an arbitrary non-builtin type.
    assert is_type_builtin_or_fake(Class) is False



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype Decidedly Object-Oriented Runtime-checking (DOOR) metaclass
hierarchy** (i.e., metaclass hierarchy driving our object-oriented type hint
class hierarchy, especially with respect to instantiation, mapping, and
memoization).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from abc import ABCMeta
from beartype.typing import (
    TYPE_CHECKING,
    # MutableMapping,
    # Union,
)
from beartype._cave._cavefast import NoneType
from beartype._data.typing.datatypingport import Hint
from beartype._util.cache.map.utilmapbig import CacheUnboundedStrong
from beartype._util.hint.utilhinttest import is_hint_cacheworthy
from threading import RLock

# If @beartype is currently being statically type-checking (e.g., by mypy),
# import the top-level "beartype" package to assist static type-checkers in
# resolving @beartype-specific forward references below.
if TYPE_CHECKING:
    import beartype

# ....................{ METACLASSES                        }....................
#FIXME: Unit test us up, please.
class _TypeHintMeta(ABCMeta):
    '''
    **Singleton abstract base class (ABC) metaclass** (i.e., the standard
    :class:`abc.ABCMeta` metaclass augmented with caching to implement the
    singleton design pattern).

    This metaclass is superior to the usual approach of implementing the
    singleton design pattern: overriding the :meth:`__new__` method of a
    singleton class to conditionally create a new instance of that class only if
    an instance has *not* already been created. Why? Because that approach
    unavoidably re-calls the :meth:`__init__` method of a previously initialized
    singleton instance on each instantiation of that class. Doing so is
    generally considered harmful.

    This metaclass instead guarantees that the :meth:`__init__` method of a
    singleton instance is only called exactly once on the first instantiation of
    that class.

    Attributes
    ----------
    __singleton : Optional[type]
        Either:

        * If the current singleton abstract base class (ABC) has been
          initialized (i.e., if the :meth:`__init__` method of this metaclass
          initializing that class with metaclass-specific logic) but a singleton
          instance of that class has *not* yet been instantiated (i.e., if the
          :meth:`__call__` method of this metaclass calling the :meth:`__new__`
          and :meth:`__init__` methods of that class (in that order) has been
          called), :data:`None`.
        * Else, the current singleton ABC has been initialized and a singleton
          instance of that class has been instantiated. In this case, that
          instance.

        For forward compatibility with future :class:`ABCMeta` changes, the name
        of this instance variable is prefixed by ``"__"`` and thus implicitly
        obfuscated by Python to be externally inaccessible.

    See Also
    --------
    https://stackoverflow.com/a/8665179/2809027
        StackOverflow answers strongly inspiring this implementation.
    '''

    # ..................{ INSTANTIATORS                      }..................
    def __call__(cls: '_TypeHintMeta', hint: Hint) -> (
        'beartype.door.TypeHint'):  # pyright: ignore
        '''
        Factory constructor magically instantiating and returning a singleton
        instance of the concrete subclass of the :class:`beartype.door.TypeHint`
        abstract base class (ABC) appropriate for handling the passed low-level
        type hint.

        Parameters
        ----------
        cls : _TypeHintMeta
            The :class:`beartype.door.TypeHint` ABC.
        hint : Hint
            Low-level type hint to be wrapped by a singleton
            :class:`beartype.door.TypeHint` instance.

        Returns
        -------
        beartype.door.TypeHint
            :class:`beartype.door.TypeHint` object wrapping this hint.

        Raises
        ------
        BeartypeDoorNonpepException
            If this class does *not* currently support the passed hint.
        BeartypeDecorHintPepSignException
            If the passed hint is *not* actually a PEP-compliant type hint.
        '''
        # print(f'!!!!!!!!!!!!! [ in _TypeHintMeta.__call__(cls={repr(cls)}, hint={repr(hint)}) ] !!!!!!!!!!!!!!!')

        # ................{ IMPORTS                            }................
        # Avoid circular import dependencies.
        from beartype.door._cls.doorsuper import TypeHint

        # ................{ TRIVIALITIES                       }................
        # If this class is a concrete subclass of the "TypeHint" abstract base
        # class (ABC) rather than ABC itself, instantiate that subclass in the
        # standard way.
        if cls is not TypeHint:
            # print('!!!!!!!!!!!!! [ _TypeHintMeta.__call__ ] instantiating subclass... !!!!!!!!!!!!!!!')
            return super().__call__(hint)
        # Else, this class is that ABC. In this case, instantiate that ABC in a
        # non-standard way.
        #
        # If this low-level type hint is already a high-level type hint wrapper,
        # return this wrapper as is. This guarantees the following constraint:
        #     >>> TypeHint(TypeHint(hint)) is TypeHint(hint)
        #     True
        elif isinstance(hint, TypeHint):
            # print('!!!!!!!!!!!!! [ _TypeHintMeta.__call__ ] reducing to noop... !!!!!!!!!!!!!!!')
            return hint
        # Else, this hint is *NOT* already a wrapper.

        # ................{ CACHING                            }................
        # Key uniquely identifying this hint, defined as either...
        hint_key = (
            # If this hint is *NOT* self-caching (i.e., *NOT* already internally
            # cached by its parent class or module), the machine-readable
            # representation of this hint. Computing this string consumes more
            # time and space and is thus performed *ONLY* where required, which
            # is for hints that are *NOT* already reduced to singleton objects.
            #
            # Note that this is *NOT* merely an optimization concern. Some
            # PEP-compliant type hints have arbitrary caller-defined and thus
            # possibly ambiguous representations. Ergo, the machine-readable
            # representation of an arbitrary hint does *NOT* uniquely identify
            # that hint in general and thus *CANNOT* be used to cache that hint.
            # Class factories producing hints with such names include:
            # * "typing.ParamSpec".
            # * "typing.TypeVar".
            repr(hint)
            if is_hint_cacheworthy(hint) else
            # Else, this hint is self-caching and thus already reduced to a
            # singleton object. In this case, the identifier identifying this
            # singleton object.
            id(hint)
        )

        #FIXME: [SPEED] Globalize this bound method as a negligible speedup.
        # Type hint wrapper wrapping this hint, efficiently cached such that
        # each hint that evaluates to the same key is wrapped by the same
        # instance of the "TypeHint" class under this Python interpreter.
        wrapper: 'beartype.door.TypeHint' = (
            _HINT_KEY_TO_WRAPPER.cache_or_get_cached_func_return_passed_arg(  # type: ignore[assignment]
                # Cache this wrapper singleton under this key.
                key=hint_key,
                # If a wrapper singleton has yet to be instantiated for this
                # hint, do so by calling this private factory method...
                value_factory=cls._make_wrapper,  # type: ignore[arg-type]
                # ...with this hint passed as the sole parameter to that method.
                arg=hint,
            ))

        # Return this wrapper.
        return wrapper

    # ..................{ PRIVATE                            }..................
    def _make_wrapper(cls: '_TypeHintMeta', hint: Hint) -> (
        'beartype.door.TypeHint'):  # pyright: ignore
        '''
        **Type hint wrapper factory** (i.e., low-level private method creating
        and returning a new :class:`beartype.door.TypeHint` instance wrapping
        the passed type hint), intended to be called by the
        :meth:`CacheUnboundedStrong.cache_or_get_cached_func_return_passed_arg`
        method to create a new type hint wrapper singleton for the passed hint.

        Parameters
        ----------
        cls : _TypeHintMeta
            The :class:`beartype.door.TypeHint` ABC.
        hint : Hint
            Low-level type hint to be wrapped by a singleton
            :class:`beartype.door.TypeHint` instance.

        Returns
        -------
        beartype.door.TypeHint
            :class:`beartype.door.TypeHint` object wrapping this hint.

        Raises
        ------
        BeartypeDoorNonpepException
            If this class does *not* currently support the passed hint.
        BeartypeDecorHintPepSignException
            If the passed hint is *not* actually a PEP-compliant type hint.
        '''

        # ................{ IMPORTS                            }................
        # Avoid circular import dependencies.
        from beartype.door._cls.util.doorclsmap import get_typehint_subclass

        # ................{ REDUCTION                          }................
        # Reduce this hint to a more amenable form suitable for mapping to a
        # concrete "TypeHint" subclass if desired.
        #
        # Note that this reduction intentionally ignores the entire
        # "beartype._check.convert" subpackage. Although submodules of that
        # subpackage do perform various coercions, reductions, and sanitizations
        # of low-level PEP-compliant type hints, they do so only for the express
        # purpose of dynamic code generation. That subpackage is *NOT*
        # general-purpose and is, in fact, harmful in this context. Why? Because
        # that subpackage erodes the semantic meaning from numerous type hints
        # that this subpackage necessarily preserves.
        #
        # ................{ REDUCTION ~ pep 484 : none         }................
        # If this is the PEP 484-compliant "None" singleton, reduce this hint to
        # the type of that singleton. While *NOT* explicitly defined by the
        # "typing" module, PEP 484 explicitly supports this singleton:
        #     When used in a type hint, the expression None is considered
        #     equivalent to type(None).
        #
        # The "None" singleton is used to type callables lacking an explicit
        # "return" statement and thus absurdly common. Ergo, detect this early.
        if hint is None:
            hint = NoneType  # type: ignore[assignment]
        # Else, this is *NOT* the PEP 484-compliant "None" singleton.

        # ................{ INSTANTIATION                      }................
        # Concrete "TypeHint" subclass handling this hint if this hint is
        # supported by an existing "TypeHint" subclass *OR* raise an exception
        # otherwise (i.e., if this hint is currently unsupported).
        wrapper_subclass = get_typehint_subclass(hint)  # pyright: ignore
        # print(f'!!!!!!!!!!!!! [ in {repr(cls)}.__new__() ] !!!!!!!!!!!!!!!')

        # Type hint wrapper wrapping this hint as a new singleton instance of
        # this subclass.
        wrapper = wrapper_subclass(hint)
        # wrapper = super(_TypeHintMeta, wrapper_subclass).__call__(hint)
        # print('!!!!!!!!!!!!! [ _TypeHintMeta.__call__ ] caching and returning singleton... !!!!!!!!!!!!!!!')

        # Return this wrapper.
        return wrapper

# ....................{ PRIVATE ~ mappings                 }....................
#FIXME: Would've been nice if this had worked, but "pyright" gonna be "pyright".
# _HINT_KEY_TO_WRAPPER_HINT = MutableMapping[
#     Union[int, str], 'beartype.door.TypeHint']
# '''
# PEP-compliant type hint matching the type hint wrapper cache defined below.
# '''
# _HINT_KEY_TO_WRAPPER: _HINT_KEY_TO_WRAPPER_HINT = CacheUnboundedStrong(  # type: ignore[assignment]


_HINT_KEY_TO_WRAPPER = CacheUnboundedStrong(
    # Prefer the slower reentrant lock type for safety. As the subpackage name
    # implies, the DOOR API is recursive and thus requires reentrancy.
    lock_type=RLock,
)
'''
**Type hint wrapper cache** (i.e., non-thread-safe cache mapping from the
machine-readable representations of all type hints to cached singleton instances
of concrete subclasses of the :class:`beartype.door.TypeHint` abstract base
class (ABC) wrapping those hints).

Design
------
**This dictionary is intentionally thread-safe.** Why? Because this dictionary
is used to ensure that :class:`beartype.door.TypeHint` instances are singletons,
enabling callers to reliably implement higher-level abstractions memoized (i.e.,
cached) against these singletons. Those abstractions could be module-scoped and
thus effectively global. To prevent race conditions between competing threads
contending over those globals, this dictionary *must* be thread-safe.

**This dictionary is intentionally designed as a naive dictionary rather than a
robust LRU cache,** for the same reasons that callables accepting hints are
memoized by the :func:`beartype._util.cache.utilcachecall.callable_cached`
rather than the :func:`functools.lru_cache` decorator. Why? Because:

* The number of different type hints instantiated across even worst-case
  codebases is negligible in comparison to the space consumed by those hints.
* The :attr:`sys.modules` dictionary persists strong references to all
  callables declared by previously imported modules. In turn, the
  ``func.__annotations__`` dunder dictionary of each such callable persists
  strong references to all type hints annotating that callable. In turn, these
  two statements imply that type hints are *never* garbage collected but
  instead persisted for the lifetime of the active Python process. Ergo,
  temporarily caching hints in an LRU cache is pointless, as there are *no*
  space savings in dropping stale references to unused hints.

**This dictionary intentionally caches machine-readable representation strings
hashes rather than alternative keys** (e.g., actual hashes). Why? Disambiguity.
Although comparatively less efficient in both space and time to construct than
hashes, the :func:`repr` strings produced for two dissimilar type hints *never*
ambiguously collide unless an external caller maliciously modified one or more
identifying dunder attributes of those hints (e.g., the ``__module__``,
``__qualname__``, and/or ``__name__`` dunder attributes). That should *never*
occur in production code. Meanwhile, the :func:`hash` values produced for two
dissimilar type hints *commonly* ambiguously collide. This is why hashable
containers (e.g., :class:`dict`, :class:`set`) explicitly handle hash table
collisions and why we are *not* going to do so.

Likewise, this dictionary intentionally caches machine-readable representations
of low-level type hints rather than those hints themselves. Since increasingly
many hints are no longer self-caching (e.g., PEP 585-compliant type hints like
"list[str]"), the latter *cannot* be guaranteed to be singletons and thus safely
used as cache keys.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **text label utilities** (i.e., low-level callables creating and
returning human-readable strings describing prominent objects or types, intended
to be embedded in human-readable error messages).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype._data.typing.datatyping import BoolTristate
from beartype._util.utilobject import (
    get_object_name,
    get_object_type_name,
)
from collections.abc import Callable

# ....................{ LABELLERS ~ object                 }....................
#FIXME: Unit test us up, please.
def label_object(
    # Mandatory parameters.
    obj: object,

    # Optional parameters.
    is_color: BoolTristate = False,
    is_context: BoolTristate = None,
) -> str:
    '''
    Human-readable label describing the passed arbitrary object.

    Parameters
    ----------
    obj : object
        Object to be labelled.
    is_color : BoolTristate, default: False
        Tri-state colouring boolean governing ANSI usage. See the
        :attr:`beartype.BeartypeConf.is_color` attribute for further details.
        Defaults to :data:`False`.
    is_context : BoolTristate, default: None
        Either:

        * :data:`True`, in which case this label is suffixed by additional
          metadata contextually disambiguating the passed object if that object
          is a callable, including:

          * The line number of the first line declaring that callable in its
            underlying source code module file.
          * The absolute filename of that file.

        * :data:`False`, in which case this label is *not* suffixed by such
          metadata.
        * :data:`None`, in which case this label is conditionally suffixed by
          such metadata only if that callable is a lambda function and thus
          ambiguously lacks any semblance of an innate context.

        Defaults to :data:`None`.

    Returns
    -------
    str
        Human-readable label describing this hintable.
    '''

    # Avoid circular import dependencies.
    from beartype._util.text.utiltextrepr import represent_object

    # If this object is a type, return a label describing this type.
    if isinstance(obj, type):
        return label_type(cls=obj, is_color=is_color)
    # Else, this beartypeable is *NOT* a type.
    #
    # If this object is a callable, return a label describing this callable.
    elif callable(obj):
        return label_callable(
            func=obj,
            is_color=is_color,
            is_context=is_context,
        )
    # Else, this object is neither a type or callable.

    # Fallback to a simplistically heuristic concatenating (in order):
    # * The human-readable kind of this object.
    # * The machine-readable representation of this object.
    return f'{label_object_kind(obj)} {represent_object(obj)}'


def label_object_kind(obj: object) -> str:
    '''
    Human-readable label describing the **kind** (i.e., single concise noun
    synopsizing the category of) of the passed object.

    Parameters
    ----------
    obj : object
        Object to describe the kind of.

    Returns
    -------
    str
        Human-readable label describing the kind of this object.
    '''

    # ....................{ IMPORTS                        }....................
    # Avoid circular import dependencies.
    from beartype._util.func.utilfunctest import (
        is_func_async,
        is_func_async_generator,
        is_func_classmethod,
        is_func_coro,
        is_func_codeobjable,
        is_func_property,
        is_func_staticmethod,
        is_func_sync_generator,
    )
    from beartype._util.func.arg.utilfuncargget import (
        get_func_arg_name_first_or_none)

    # ....................{ LOCALS                         }....................
    # String to be returned, defaulting to a sensible placeholder.
    obj_label = 'object'

    # ....................{ TYPE                           }....................
    # If this object is a pure-Python class, return an appropriate string.
    if isinstance(obj, type):
        obj_label = 'class'
    # Else, this object is *NOT* a pure-Python class.
    #
    # ....................{ CODEOBJABLE                    }....................
    # If this object is (or encapsulates) a pure-Python callable...
    elif is_func_codeobjable(obj):
        # Human-readable prefix describing the exotic nature of this callable if
        # this is callable is exotic (e.g., coroutine or generator factory)
        # suffixed by trailing whitespace *OR* the empty string otherwise.
        func_prefix = ''

        # Human-readable suffix describing the general nature of this callable
        # (e.g., function, method) suffixed by trailing whitespace.
        func_suffix = ''

        # If this object is an asynchronous callable factory...
        if is_func_async(obj):
            # If this object is a coroutine factory, use an appropriate prefix.
            if is_func_coro(obj):
                func_prefix = 'coroutine factory '
            # If this object is an asynchronous generator factory, use an
            # appropriate prefix.
            elif is_func_async_generator(obj):
                func_prefix = 'asynchronous generator factory '
            # Else, this object is an unrecognized kind of asynchronous callable
            # factory. In this case, fallback to a generic prefix.
            #
            # Note that this should *NEVER* occur. Since *ALL* asynchronous
            # callable factories are either coroutine or asynchronous generator
            # factories, one of the above conditional branches should have been
            # entered instead. Nonetheless, preparation prevents disasters.
            else:  # pragma: no cover
                func_prefix = 'asynchronous '
        # Else, this object is a synchronous callable.
        #
        # If this object is an synchronous generator factory, use an appropriate
        # prefix.
        elif is_func_sync_generator(obj):
            func_prefix = 'generator factory '
        # Else, this object is a standard synchronous callable. In this case,
        # avoid prefixing this callable by a leading substring.

        # Name of the first parameter accepted by that callable if any *OR*
        # "None" otherwise (i.e., if that callable is argumentless).
        arg_first_name = get_func_arg_name_first_or_none(
            func=obj,
            # If that callable is a C-based bound method descriptor
            # encapsulating a pure-Python instance or class method, instruct
            # this getter to return the name of the first mandatory flexible
            # parameter accepted by that method. In general, this parameter is
            # semantically ignorable and indeed *MUST* be ignored. In this case,
            # however, this name conveys (marginally) meaningful syntax under
            # the safe assumption that author of that callable adheres to PEP 8
            # community standards and is thus sane rather than insane. Notably,
            # this name is then guaranteed to be either:
            # * "self" if that descriptor encapsulates an instance method.
            # * "cls" if that descriptor encapsulates a class method.
            is_omit_boundmethod_arg_first=False,
        )

        # If this is the canonical first "self" parameter typically accepted by
        # instance methods, assume this to be an instance method.
        #
        # Note that this heuristic fails in uncommon edge cases -- but that
        # that's largely irrelevant here. This function is *ONLY* intended to
        # generate human-readable exception and warning messages. Since this is
        # hardly mission-critical, false positives are reluctantly acceptable.
        if arg_first_name == 'self':
            func_suffix = 'method'
        # Else, this is *NOT* the canonical first "self" parameter.
        #
        # If this is the canonical first "cls" parameter typically accepted by
        # class methods, assume this to be a class method.
        elif arg_first_name == 'cls':
            func_suffix = 'class method'
        # Else, this is neither the canonical first "self" nor "cls" parameter.
        # In this case, assume this to be a non-method callable.
        else:
            func_suffix = 'function'

        # Return the concatenation of these substrings.
        # print(f'func_prefix: {func_prefix}; func_suffix: {func_suffix}')
        obj_label = f'{func_prefix}{func_suffix}'
    # Else, this object is *NOT* a pure-Python callable.
    #
    # ....................{ DESCRIPTOR                     }....................
    # If this object is a C-based uncallable class method descriptor, return an
    # appropriate string.
    elif is_func_classmethod(obj):
        obj_label = 'class method descriptor'
    # Else, this object is *NOT* a C-based uncallable class method descriptor.
    #
    # If this object is a C-based uncallable property method descriptor, return
    # an appropriate string.
    elif is_func_property(obj):
        obj_label = 'property method descriptor'
    # Else, this object is *NOT* a C-based uncallable property method
    # descriptor.
    #
    # If this object is a C-based uncallable static method descriptor, return
    # an appropriate string.
    elif is_func_staticmethod(obj):
        obj_label = 'static method descriptor'
    # Else, this object is *NOT* a C-based uncallable static method descriptor.

    # ....................{ RETURN                         }....................
    # Return this label.
    return obj_label


#FIXME: Unit test us up, please.
def label_object_context(obj: object) -> str:
    '''
    Human-readable label describing the **context** (i.e., absolute filename of
    the module or script physically declaring the passed object *and* the
    1-based line number of the first line declaring this object in this file) of
    this object if this object is either a callable or class declared on-disk
    *or* the empty string otherwise (i.e., if this object is neither a callable
    nor class *or* is either a callable or class declared in-memory).

    Parameters
    ----------
    func : object
        Object to label the context of.

    Returns
    -------
    str
        Human-readable label describing the context of this object.
    '''

    # Defer test-specific imports.
    from beartype._util.utilobject import get_object_filename_or_none
    from beartype._util.module.utilmodget import (
        get_object_module_line_number_begin)

    # Absolute filename of the module or script physically declaring this object
    # if this object was defined on-disk *OR* "None" otherwise (i.e., if this
    # object was defined in-memory).
    obj_filename = get_object_filename_or_none(obj)

    # If this object is defined on-disk...
    if obj_filename:
        # Line number of the first line declaring this object in that file.
        obj_lineno = get_object_module_line_number_begin(obj)

        # Return a string describing the context of this object.
        return f'in file "{obj_filename}" line {obj_lineno}'
    # Else, this object was defined in-memory. In this case, avoid attempting to
    # needlessly contextualize this object.

    # Let's hear it for giving up here and going home. Yeah! Go, @beartype!
    return ''

# ....................{ LABELLERS ~ callable               }....................
#FIXME: Unit test up the "is_context" parameter, which is currently untested.
def label_callable(
    # Mandatory parameters.
    func: Callable,

    # Optional parameters.
    is_color: BoolTristate = False,
    is_context: BoolTristate = None,
) -> str:
    '''
    Human-readable label describing the passed **callable** (e.g., function,
    method, property).

    Parameters
    ----------
    func : Callable
        Callable to be labelled.
    is_color : BoolTristate, default: False
        Tri-state colouring boolean governing ANSI usage. See the
        :attr:`beartype.BeartypeConf.is_color` attribute for further details.
        Defaults to :data:`False`.
    is_context : BoolTristate, default: None
        Either:

        * :data:`True`, in which case this label is suffixed by additional
          metadata contextually disambiguating that callable, including:

          * The line number of the first line declaring that callable in its
            underlying source code module file.
          * The absolute filename of that file.

        * :data:`False`, in which case this label is *not* suffixed by such
          metadata.
        * :data:`None`, in which case this label is conditionally suffixed by
          such metadata only if that callable is a lambda function and thus
          ambiguously lacks any semblance of an innate context.

        Defaults to :data:`None`.

    Returns
    -------
    str
        Human-readable label describing this callable.
    '''
    assert callable(func), f'{repr(func)} uncallable.'
    assert isinstance(is_context, bool) or is_context is None, (  # <-- "NoneTypeOr" is unavailable here
        f'{repr(is_context)} not tri-state boolean.')

    # Avoid circular import dependencies.
    from beartype._util.func.arg.utilfuncarglen import (
        get_func_args_flexible_len)
    from beartype._util.func.utilfunccodeobj import get_func_codeobj
    from beartype._util.func.utilfunctest import is_func_lambda
    from beartype._util.text.utiltextansi import color_attr_name

    # Substring prefixing the string to be returned, typically identifying the
    # specialized type of that callable if that callable has a specialized type.
    func_label_prefix = ''

    # Fully-qualified name of that callable, coloured if requested.
    func_label = color_attr_name(
        text=f' {get_object_name(func)}()', is_color=is_color)

    # Substring suffixing the string to be returned, typically contextualizing
    # that callable with respect to its on-disk code module file.
    func_label_suffix = ''

    #FIXME: *HMM.* This branch should almost certainly be folded into the
    #existing label_object_kind() function, which would then dramatically
    #simplify this logic here. Let's do this, yo!
    # If the passed callable is a pure-Python lambda function, that callable
    # has *NO* unique fully-qualified name. In this case, return a string
    # uniquely identifying this lambda from various code object metadata.
    if is_func_lambda(func):
        # Code object underlying this lambda.
        func_codeobj = get_func_codeobj(func)

        # Substring preceding the string to be returned.
        func_label_prefix = (
            f'lambda function accepting '
            f'{get_func_args_flexible_len(func_codeobj)} argument(s)'
        )

        # If the caller failed to request an explicit contextualization, default
        # to contextualizing this lambda function.
        if is_context is None:
            is_context = True
        # Else, the caller requested an explicit contextualization. In this
        # case, preserve that contextualization as is.
    # Else, the passed callable is *NOT* a pure-Python lambda function and thus
    # has a unique fully-qualified name. In this case, prefix this label with a
    # substring describing the kind of that callable.
    else:
        func_label_prefix = label_object_kind(func)

    # If contextualizing that callable, just do it already. Go, @beartype! Go!
    if is_context:
        func_label_suffix = f' {label_object_context(func)}'
    # Else, we are *NOT* contextualizing that callable.

    # Return that prefix followed by the fully-qualified name of that callable.
    return f'{func_label_prefix}{func_label}{func_label_suffix}'

# ....................{ LABELLERS ~ exception              }....................
def label_exception(exception: Exception) -> str:
    '''
    Human-readable label describing the passed exception.

    Caveats
    -------
    **The label returned by this function does not describe the traceback
    originating this exception.** To do so, consider calling the standard
    :func:`traceback.format_exc` function instead.

    Parameters
    ----------
    exception : Exception
        Exception to be labelled.

    Returns
    -------
    str
        Human-readable label describing this exception.
    '''
    assert isinstance(exception, Exception), (
        f'{repr(exception)} not exception.')

    # Return the fully-qualified name of the class of this exception followed by
    # this exception's message.
    return f'{get_object_type_name(exception)}: {str(exception)}'

# ....................{ LABELLERS ~ pith                   }....................
def label_pith_value(
    # Mandatory parameters.
    pith: object,

    # Optional parameters.
    is_color: BoolTristate = False,
) -> str:
    '''
    Human-readable label describing the passed value of the **current pith**
    (i.e., arbitrary object violating the current type check) *not* suffixed by
    delimiting whitespace.

    Parameters
    ----------
    pith : object
        Arbitrary object violating the current type check.
    is_color : BoolTristate
        Tri-state colouring boolean governing ANSI usage. See the
        :attr:`beartype.BeartypeConf.is_color` attribute for further details.
        Defaults to :data:`False`.

    Returns
    -------
    str
        Human-readable label describing this pith value.
    '''

    # Avoid circular import dependencies.
    from beartype._util.text.utiltextansi import color_pith
    from beartype._util.text.utiltextrepr import represent_object

    # Glory be to the one liner that you are about to read.
    return color_pith(text=represent_object(pith), is_color=is_color)

# ....................{ LABELLERS ~ type                   }....................
def label_type(
    # Mandatory parameters.
    cls: type,

    # Optional parameters.
    is_color: BoolTristate = False,
) -> str:
    '''
    Human-readable label describing the passed class.

    Parameters
    ----------
    cls : type
        Class to be labelled.
    is_color : BoolTristate, optional
        Tri-state colouring boolean governing ANSI usage. See the
        :attr:`beartype.BeartypeConf.is_color` attribute for further details.
        Defaults to :data:`False`.

    Returns
    -------
    str
        Human-readable label describing this class.
    '''
    assert isinstance(cls, type), f'{repr(cls)} not class.'

    # Avoid circular import dependencies.
    from beartype._util.cls.utilclstest import is_type_builtin
    from beartype._util.hint.pep.proposal.pep544 import (
        is_hint_pep544_protocol)
    from beartype._util.hint.pep.proposal.pep484585.generic.pep484585gentest import (
        is_hint_pep484585_generic_subbed)
    from beartype._util.text.utiltextansi import color_type

    # Label to be returned.
    classname: str = None  # type: ignore[assignment]

    # If this type is a PEP 484- or 585-compliant subscripted generic, the best
    # label is simply the machine-readable representation of this generic (e.g.,
    # the string "MuhGeneric[str]" given the generic type declaration "class
    # MuhGeneric[T](typing.Generic[T]):"). Alternative labels defined below fail
    # to include the child hint subscripting this subscripted generic.
    if is_hint_pep484585_generic_subbed(cls):  # pyright: ignore
        classname = repr(cls)
    # Else, this type is *NOT* a PEP 484- or 585-compliant subscripted generic.
    # In this case...
    else:
        # Fully-qualified name of this type.
        classname = get_object_type_name(cls)
        # print(f'cls {cls} classname: {classname}')

        # If this name contains *NO* periods, this class is actually a builtin
        # type (e.g., "list"). Since builtin types are well-known and thus
        # self-explanatory, this name requires no additional labelling. In this
        # case, return this name as is.
        if '.' not in classname:
            pass
        # Else, this name contains one or more periods but could still be a
        # builtin indirectly accessed via the standard "builtins" module.
        #
        # If this name is that of a builtin type uselessly prefixed by the name
        # of the module declaring all builtin types (e.g., "builtins.list"),
        # reduce this name to the unqualified basename of this type (e.g.,
        # "list").
        elif is_type_builtin(cls):
            classname = cls.__name__
        # Else, this is a non-builtin class. Non-builtin classes are *NOT*
        # well-known and thus benefit from additional labelling.
        #
        # If this class is a PEP 544-compliant protocol supporting structural
        # subtyping, label this protocol.
        elif is_hint_pep544_protocol(cls):
            # print(f'cls {cls} is protocol!')
            classname = f'<protocol "{classname}">'
        # Else if this class is a standard abstract base class (ABC) defined by
        # a standard submodule also known to support structural subtyping (e.g.,
        # "collections.abc.Hashable", "contextlib.AbstractContextManager"),
        # label this ABC as a protocol.
        #
        # Note that user-defined ABCs do *NOT* generally support structural
        # subtyping. Doing so requires esoteric knowledge of undocumented and
        # mostly private "abc.ABCMeta" metaclass internals unlikely to be
        # implemented by third-party developers. Thanks to the lack of both
        # publicity and standardization, there exists *NO* general-purpose means
        # of detecting whether an arbitrary class supports structural subtyping.
        elif (
            classname.startswith('collections.abc.') or
            classname.startswith('contextlib.')
        ):
            classname = f'<protocol ABC "{classname}">'
        # Else, this is a standard class. In this case, label this class as such.
        else:
            classname = f'<class "{classname}">'

    # Return this labelled classname, possibly coloured.
    return color_type(text=classname, is_color=is_color)


#FIXME: Silly, honestly. Is this still called anywhere? Simplify to trivial
#calls to the vastly superior label_type() function defined above, please.
def label_object_type(obj: object) -> str:
    '''
    Human-readable label describing the class of the passed object.

    Parameters
    ----------
    obj : object
        Object whose class is to be labelled.

    Returns
    -------
    str
        Human-readable label describing the class of this object.
    '''

    # Tell me why, why, why I curse the sky! ...no, srsly.
    return label_type(type(obj))



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

"""
**Beartype object label utility unit tests.**

This submodule unit tests the public API of the private
:mod:`beartype._util.text.utiltextlabel` submodule.
"""


# ....................{ IMPORTS                            }....................
# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS                              }....................
def test_label_object_kind() -> None:
    '''
    Test the :func:`beartype._util.text.utiltextlabel.label_object_kind`
    function.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype._util.text.utiltextlabel import label_object_kind
    from beartype_test.a00_unit.data.data_type import (
        ClassCallable,
        Class,
        async_coroutine_factory,
        async_generator_factory,
        function,
        sync_generator_factory,
    )
    from beartype_test.a00_unit.data.func.data_func import (
        func_args_3_flex_mandatory_optional_varkw)

    # ....................{ LOCALS                         }....................
    # Tuple of 2-tuples "(beartypeable, kind)", where:
    # * "beartypeable" is an input object to be passed to this labeller.
    # * "kind" is the string expected to be returned from this labeller when
    #   passed this input object.
    BEARTYPEABLES_KINDS = (
        # Builtin type.
        (str, 'class'),

        # User-defined class.
        (Class, 'class'),

        # Pure-Python argumentless function.
        (function, 'function'),

        # Pure-Python argumentative function.
        (func_args_3_flex_mandatory_optional_varkw, 'function'),

        # Pure-Python instance method.
        (Class.instance_method, 'method'),

        # Pure-Python class method.
        (Class.class_method, 'class method'),

        # Pure-Python coroutine factory function.
        (async_coroutine_factory, 'coroutine factory function'),

        # Pure-Python asynchronous generator factory function.
        (async_generator_factory, 'asynchronous generator factory function'),

        # Pure-Python asynchronous generator factory function.
        (sync_generator_factory, 'generator factory function'),

        # Object that is neither a pure-Python class, function, *NOR* method. In
        # this case, pass a pseudo-callable (i.e., object whose class defines
        # the __call__() dunder method) to exercise a *POSSIBLE* edge case.
        (ClassCallable(), 'object'),
    )

    # ....................{ PASS                           }....................
    # For each such input object and expected string...
    for beartypeable, kind in BEARTYPEABLES_KINDS:
        # Assert this labeller returns the expected string when passed this
        # input object.
        assert label_object_kind(beartypeable) == kind


def test_label_callable() -> None:
    '''
    Test the :func:`beartype._util.text.utiltextlabel.label_callable`
    function.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype._util.text.utiltextlabel import label_callable
    from beartype_test.a00_unit.data.data_type import (
        async_coroutine_factory,
        async_generator_factory,
        sync_generator_factory,
    )
    from beartype_test.a00_unit.data.util.mod.data_utilmodule_line import (
        like_snakes_that_watch_their_prey,
        ozymandias,
        which_yet_survive,
    )

    # ....................{ PASS                           }....................
    # Assert this labeller labels an on-disk lambda function as expected.
    two_vast_and_trunkless_legs_of_stone = label_callable(ozymandias)
    assert isinstance(two_vast_and_trunkless_legs_of_stone, str)
    assert 'lambda' in two_vast_and_trunkless_legs_of_stone

    # Assert this labeller labels an in-memory lambda function as expected.
    the_hand_that_mocked_them = label_callable(which_yet_survive)
    assert isinstance(the_hand_that_mocked_them, str)
    assert 'lambda' in the_hand_that_mocked_them

    # Assert this labeller labels an on-disk non-lambda callable as expected.
    tell_that_its_sculptor_well_those_passions_read = label_callable(
        like_snakes_that_watch_their_prey)
    assert isinstance(tell_that_its_sculptor_well_those_passions_read, str)
    assert like_snakes_that_watch_their_prey.__name__ in (
        tell_that_its_sculptor_well_those_passions_read)

    # Assert this labeller labels an on-disk coroutine as expected.
    async_coroutine_label = label_callable(async_coroutine_factory)
    assert isinstance(async_coroutine_label, str)
    assert async_coroutine_factory.__name__ in async_coroutine_label
    assert 'coroutine' in async_coroutine_label

    # Assert this labeller labels an on-disk asynchronous generator as expected.
    async_generator_label = label_callable(async_generator_factory)
    assert isinstance(async_generator_label, str)
    assert async_generator_factory.__name__ in async_generator_label
    assert 'asynchronous generator' in async_generator_label

    # Assert this labeller labels an on-disk synchronous generator as expected.
    sync_generator_label = label_callable(sync_generator_factory)
    assert isinstance(sync_generator_label, str)
    assert sync_generator_factory.__name__ in sync_generator_label
    assert 'generator' in sync_generator_label


def test_label_type() -> None:
    '''
    Test the :func:`beartype._util.text.utiltextlabel.label_type`
    function.
    '''

    # Defer test-specific imports.
    from beartype._util.text.utiltextlabel import label_type
    from beartype_test.a00_unit.data.data_type import Class

    # Assert this labeller returns the expected label for a builtin type.
    assert label_type(str) == 'str'

    # Assert this labeller returns the expected label for the non-builtin type
    # of the "None" singleton, exercising a common edge case.
    assert label_type(type(None)) == '<class "builtins.NoneType">'

    # Assert this labeller returns the expected label for a user-defined type.
    assert label_type(Class) == (
        '<class "beartype_test.a00_unit.data.data_type.Class">')



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype decorator** :pep:`692`-compliant ``**kwargs:
typing.Unpack[typing.TypedDict]`` unit tests.

This submodule unit tests :pep:`692` support implemented in the
:func:`beartype.beartype` decorator.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from beartype_test._util.mark.pytskip import skip_if_python_version_less_than

# ....................{ TESTS                              }....................
@skip_if_python_version_less_than('3.12.0')
def test_decor_pep692() -> None:
    '''
    Test :pep:`692` support implemented in the :func:`beartype.beartype`
    decorator if the active Python interpreter targets Python >= 3.12 *or* skip
    otherwise.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype import beartype
    from beartype.roar import BeartypeDecorHintPep646692Exception
    from beartype.typing import (
        Tuple,
        TypedDict,
    )
    from beartype._util.api.standard.utiltyping import get_typing_attrs
    from pytest import raises

    # ....................{ CLASSES                        }....................
    @beartype
    class BlueCavernMould(TypedDict):
        '''
        Typed dictionary annotating one or more arbitrary key-value pairs.
        '''

        nurses_of_rainbow_flowers: str
        and_branching_moss: bytes

    # ....................{ FACTORIES                      }....................
    # For the "Unpack" type hint factory exported by each typing module...
    for Unpack in get_typing_attrs('Unpack'):
        # ....................{ CALLABLES                  }....................
        @beartype
        def commit_the_colours(
            **kwargs: Unpack[BlueCavernMould]) -> str:
            '''
            Arbitrary callable decorated by :func:`beartype.beartype` accepting
            a variadic keyword parameter annotated by as a :pep:`692`-compliant
            unpacked typed dictionary subclass.
            '''

            # Return the string concatenation of these keyword arguments.
            return (
                kwargs['nurses_of_rainbow_flowers'] +
                kwargs['and_branching_moss'].decode('utf-8')
            )

        # ....................{ PASS                       }....................
        # Assert that this callable returns a tuple of all passed positional
        # parameters.
        assert commit_the_colours(
            nurses_of_rainbow_flowers='Commit the colours of ',
            and_branching_moss=b'that varying cheek,',
        ) == 'Commit the colours of that varying cheek,'

        # ....................{ FAIL                       }....................
        #FIXME: Additionally assert that calling the above commit_the_colours()
        #function with unrecognized or invalid keyword arguments raises the
        #expected type-checking violation *AFTER* we deeply support PEP 692.

        # Assert that @beartype raises the expected exception when decorating a
        # callable accepting a variadic keyword parameter annotated by a hint
        # unpacking a PEP 692-noncompliant object (i.e., *ANY* object other than
        # a PEP 692-compliant typed dictionary subclass).
        with raises(BeartypeDecorHintPep646692Exception):
            @beartype
            def that_snowy_breast(
                **those_dark_and_drooping_eyes: Unpack[Tuple[str]]) -> str:
                return next(iter(those_dark_and_drooping_eyes.keys()))



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype :pep:`3119` unit tests.

This submodule unit tests :pep:`3119` support implemented in the
:func:`beartype.beartype` decorator.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS                              }....................
def test_decor_pep3119() -> None:
    '''
    Test :pep:`3119` support implemented in the :func:`beartype.beartype`
    decorator -- particularly with respect to **partially initialized
    metaclasses** (i.e., metaclasses with ``__instancecheck__()` and/or
    ``__subclasscheck__()`` dunder methods that are *not* safely callable at the
    early decoration time that :func:`beartype.beartype` attempts to call those
    methods).
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype import beartype
    from beartype.roar import BeartypeDecorHintPep3119Exception
    from beartype.typing import (
        Tuple,
        Type,
    )
    from pytest import raises

    # Intentionally import from the "typing" module to allow the definition of
    # an uncheckable protocol below.
    from typing import Protocol

    # ....................{ CLASSES                        }....................
    class Pep3119Metaclass(type):
        '''
        Arbitrary :pep:`3119`-compliant metaclass guaranteed to be only
        partially initialized at the early decoration time that
        :func:`beartype.beartype` attempts to call the :meth:`__instancecheck__`
        and :meth:`__subclasscheck__` dunder methods defined below.
        '''

        def __instancecheck__(cls: type, other: object) -> bool:
            return each_like_a_corpse_within_its_grave()

        def __subclasscheck__(cls: type, other: type) -> bool:
            return each_like_a_corpse_within_its_grave()


    class Pep3119Class(metaclass=Pep3119Metaclass):
        '''
        Arbitrary class whose metaclass is a :pep:`3119`-compliant metaclass
        guaranteed to be only partially initialized.
        '''

        pass


    class NonisinstanceableProtocol(Protocol):
        '''
        Arbitrary **non-isinstanceable protocol** (i.e., protocol *not*
        decorated by the :func:`typing.runtime_checkable` decorator and
        hence *not* checkable at runtime, as doing so raises a
        :exc:`TypeError` from the ``__instancecheck__`` dunder method of the
        metaclass of this protocol).
        '''

        pass

    # ....................{ FUNCTIONS                      }....................
    # Implicitly assert that decorating a function annotated by a partially
    # initialized PEP 3119-compliant metaclass does *NOT* raise an exception.
    @beartype
    def winged_seeds(
        # Implicitly invoke the __instancecheck__() method defined above.
        where_they_lie_cold_and_low: Pep3119Class,
        # Implicitly invoke the __subclasscheck__() method defined above.
        driving_sweet_buds_like_flocks_to_feed_in_air: Type[Pep3119Class],
    ) -> Tuple[Pep3119Class, Type[Pep3119Class]]:
        '''
        Arbitrary function annotated as accepting a parameter that is an
        instance of an arbitrary class whose metaclass is a
        :pep:`3119`-compliant metaclass guaranteed to be only partially
        initialized at this early decoration time.
        '''

        return (
            where_they_lie_cold_and_low,
            driving_sweet_buds_like_flocks_to_feed_in_air,
        )


    def each_like_a_corpse_within_its_grave() -> bool:
        '''
        Arbitrary boolean function.
        '''

        return True

    # ....................{ PASS                           }....................
    # Arbitrary instance of this class.
    thine_azure_sister_of_the_Spring_shall_blow = Pep3119Class()

    # Assert that calling the decorated function defined above succeeds.
    assert winged_seeds(
        thine_azure_sister_of_the_Spring_shall_blow, Pep3119Class) == (
        thine_azure_sister_of_the_Spring_shall_blow, Pep3119Class)

    # ....................{ FAIL                           }....................
    # Assert that decorating a function annotated as accepting a parameter
    # hinted as an instance of a non-isinstanceable class raises the expected
    # exception.
    with raises(BeartypeDecorHintPep3119Exception):
        @beartype
        def her_clarion_over_the_dreaming_earth(
            and_lie: NonisinstanceableProtocol) -> None:
            pass

    # Assert that decorating a function annotated as accepting a parameter
    # hinted as a subclass of a non-issubclassable class raises the expected
    # exception.
    with raises(BeartypeDecorHintPep3119Exception):
        @beartype
        def with_living_hues_and_odours_plain(
            and_hill: Type[NonisinstanceableProtocol]) -> None:
            pass



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`484`-compliant type hint utilities.

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
# Intentionally import PEP 484-compliant "typing" type hint factories rather
# than possibly PEP 585-compliant "beartype.typing" type hint factories.
from typing import Tuple

# ....................{ HINTS                              }....................
#FIXME: This very much does *NOT* belong here.
HINT_PEP484_TUPLE_EMPTY = Tuple[()]
'''
:pep:`484`-compliant empty fixed-length tuple type hint.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **code object globals** (i.e., global constants describing code
objects of callables, classes, and modules).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ STRINGS                            }....................
FUNC_CODEOBJ_NAME_MODULE = '<module>'
'''
Arbitrary string constant unconditionally assigned to the ``co_name`` instance
variables of the code objects of all pure-Python modules (i.e., the top-most
lexical scope of each module in the current call stack).

This constant enables callers to reliably differentiate between code objects
encapsulating:

* Module scopes, whose ``co_name`` variable is this constant.
* Callable scopes, whose ``co_name`` variable is *not* this constant.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :mod:`xarray` integration tests.

This submodule functionally tests the :mod:`beartype` package against the
third-party :mod:`xarray` package.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from beartype_test._util.mark.pytskip import skip_unless_package

# ....................{ TESTS                              }....................
@skip_unless_package('xarray')
def test_xarray_dataset() -> None:
    '''
    Integration test validating that the :mod:`beartype` package raises *no*
    unexpected exceptions when type-checking instances of the third-party
    :class:`xarray.Dataset` class known to be hostile to runtime type-checking.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype import beartype
    from beartype.roar import BeartypeCallHintParamViolation
    from pytest import raises
    from xarray import Dataset

    # ....................{ LOCALS                         }....................
    # Arbitrary xarray dataset.
    xarray_dataset = Dataset({
        'temperature': (
            ['x', 'y'], [
                [20, 25],
                [30, 35],
            ],
        ),
    })

    # ....................{ CALLABLES                      }....................
    @beartype
    def crunch_data(data: Dataset) -> Dataset:
        '''
        Arbitrary callable both accepting and returning a
        :class:`xarray.Dataset` object.
        '''

        # Do it, @beartype. Do it for @leycec!
        return data

    # ....................{ PASS                           }....................
    # Implicitly assert that a @beartype-decorated callable accepts this dataset
    # *WITHOUT* raising unexpected exceptions.
    crunch_data(xarray_dataset)

    # ....................{ FAIL                           }....................
    # Assert that a @beartype-decorated callable raises the expected exception
    # when passed a non-dataset.
    with raises(BeartypeCallHintParamViolation):
        crunch_data("It's 110°F and I can feel my face peeling off my face.")



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **string singletons** (i.e., strings and data structures of strings
commonly required throughout this codebase, reducing space and time consumption
by preallocating widely used string-centric objects).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from string import punctuation

# ....................{ SETS ~ punctuation                 }....................
CHARS_PUNCTUATION = frozenset(punctuation)
'''
Frozen set of all **ASCII punctuation characters** (i.e., non-Unicode
characters satisfying the conventional definition of English punctuation).

Note that the :attr:`string.punctuation` object is actually an inefficient
string of these characters rather than an efficient collection. Ergo, this set
should *ALWAYS* be accessed in lieu of that string.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`560`-compliant **generic type hint iterators** (i.e.,
low-level callables generically iterating over the ``__orig_bases__`` dunder
attribute defining the pseudo-superclasses of :pep:`484`- and
:pep:`560`-compliant generic class hierarchies).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeDecorHintPep560Exception
from beartype._data.typing.datatypingport import (
    Hint,
    IterableHints,
)
from beartype._data.typing.datatyping import TypeException
from beartype._util.cache.pool.utilcachepoollistfixed import (
    FIXED_LIST_SIZE_MEDIUM,
    acquire_fixed_list,
    release_fixed_list,
)

# ....................{ TESTERS                            }....................
def is_hint_pep560(hint: Hint) -> bool:
    '''
    :data:`True` only if the passed hint is :pep:`560`-compliant (i.e., defines
    the ``__orig_bases__`` dunder attribute), strongly implying this hint to be
    a :pep:`560`-compliant generic.

    Parameters
    ----------
    hint : Hint
        Type hint to be inspected.

    Returns
    -------
    bool
        :data:`True` only if this hint is :pep:`560`-compliant.
    '''

    # Return true only if this hint defines this dunder attribute.
    return hasattr(hint, '__orig_bases__')

# ....................{ ITERATORS                          }....................
#FIXME: Unit test us up, please.
def iter_hint_pep560_bases_unerased(
    # Mandatory parameters.
    hint: Hint,

    # Optional parameters.
    exception_cls: TypeException = BeartypeDecorHintPep560Exception,
    exception_prefix: str = '',
) -> IterableHints:
    '''
    Generator iteratively yielding the one or more **unerased
    pseudo-superclasses** (i.e., unignorable PEP-compliant type hints originally
    declared as transitive superclasses prior to type erasure) of the passed
    **generic** (i.e., :pep:`484`- or :pep:`585`-compliant subscriptable type in
    either unsubscripted or subscripted form), effectively performing a
    breadth-first search (BFS) over these pseudo-superclasses.

    This generator yields the full tree of all pseudo-superclasses by
    transitively visiting both all direct pseudo-superclasses of this generic
    *and* all indirect pseudo-superclasses transitively superclassing all direct
    pseudo-superclasses of this generic. For efficiency, this generator is
    internally implemented with an efficient imperative First In First Out
    (FILO) queue rather than an inefficient (and dangerous, due to both
    unavoidable stack exhaustion and avoidable infinite recursion) tree of
    recursive function calls.

    This generator intentionally defers to *no* other :pep:`484`- or
    :pep:`585`-compliant generic functionality, enabling this generator to be
    safely called by that functionality without concern for circular logic.

    This generator is intentionally *not* memoized (e.g., by the
    ``callable_cached`` decorator). Memoization is the caller's responsibility.

    Caveats
    -------
    **The higher-level**
    :func:`beartype._check.pep.get_hint_pep484585_generic_unsubbed_bases_unerased`
    **generator should typically be called instead.** This lower-level generator
    is *only* intended to be called by equally low-level utility functions
    throughout the :mod:`beartype._util.hint.pep.proposal` subpackage.

    **This generator exhibits** :math:`O(n)` **linear time complexity for**
    :math:`n` the number of transitive pseudo-superclasses of this generic. So,
    this generator is slow. The caller is expected to memoize *all* calls to
    this generator, which is itself *not* memoized.

    Parameters
    ----------
    hint : Hint
        Generic to be inspected.
    exception_cls : TypeException, default: BeartypeDecorHintPep560Exception
        Type of exception to be raised in the event of fatal error. Defaults to
        :exc:`.BeartypeDecorHintPep560Exception`.
    exception_prefix : str, default: ''
        Human-readable substring prefixing raised exception messages. Defaults
        to the empty string.

    Yields
    ------
    Hint
        Unerased transitive pseudo-superclass originally declared as a
        superclass prior to its type erasures by this generic.

    Raises
    ------
    exception_cls
        If this hint is *not* a generic.
    '''

    # ....................{ LOCALS ~ bases : direct        }....................
    # Tuple of the one or more unerased pseudo-superclasses (i.e., originally
    # listed as superclasses prior to type erasure) of the passed generic if
    # this generic is actually a PEP 560-compliant generic *OR* "None".
    hint_bases_direct = getattr(hint, '__orig_bases__', None)
    # print(f'generic {hint} hint_bases_direct: {hint_bases_direct}')

    # If this generic is *NOT* actually a PEP 560-compliant generic...
    if hint_bases_direct is None:
        assert isinstance(exception_cls, type), (
            f'{repr(exception_cls)} not exception type.')
        assert isinstance(exception_prefix, str), (
            f'{repr(exception_prefix)} not string.')

        # Raise an exception.
        raise exception_cls(
            f'{exception_prefix}type hint {repr(hint)} '
            f'not PEP 484- or 585-compliant generic.'
        )
    # Else, this generic is a PEP 560-compliant generic.

    # ....................{ LOCALS ~ bases : indirect      }....................
    # Fixed list of the one or more unerased transitive pseudo-superclasses
    # originally listed as superclasses prior to their type erasure by this
    # generic that have yet to be visited by the breadth-first search (BFS) over
    # these pseudo-superclasses performed below.
    hint_bases = acquire_fixed_list(size=FIXED_LIST_SIZE_MEDIUM)

    # 0-based index of the currently visited pseudo-superclass of this list.
    hint_bases_index_curr = 0

    # 0-based index of one *PAST* the last pseudo-superclass of this list.
    hint_bases_index_past_last = len(hint_bases_direct)

    # Initialize this list to these direct pseudo-superclasses of this generic.
    hint_bases[0:hint_bases_index_past_last] = hint_bases_direct

    # ....................{ SEARCH                         }....................
    # While the 0-based index of the next visited pseudo-superclass does *NOT*
    # exceed that of the last pseudo-superclass in this list, there remains one
    # or more pseudo-superclasses to be visited in this BFS. Let us do so.
    while hint_bases_index_curr < hint_bases_index_past_last:
        # ....................{ LOCALS                     }....................
        # Currently visited transitive pseudo-superclass of the passed generic.
        hint_base = hint_bases[hint_bases_index_curr]

        # Yield this pseudo-superclass to the caller.
        yield hint_base

        # Tuple of the one or more unerased pseudo-superclasses of this
        # pseudo-superclass if this pseudo-superclass is itself a PEP
        # 560-compliant generic *OR* "None" otherwise (i.e., if this
        # pseudo-superclass is *NOT* a PEP 560-compliant generic).
        hint_base_bases_direct = getattr(hint_base, '__orig_bases__', None)

        # If this pseudo-superclass is a PEP 560-compliant generic...
        if hint_base_bases_direct is not None:
            # 0-based index of the last pseudo-superclass of this list
            # *BEFORE* adding onto this list.
            hint_bases_index_past_last_prev = hint_bases_index_past_last

            # 0-based index of the last pseudo-superclass of this list
            # *AFTER* adding onto this list.
            hint_bases_index_past_last += len(hint_base_bases_direct)

            # Enqueue all parent pseudo-superclasses of this child
            # pseudo-superclass onto this list for later visitation by this BFS.
            hint_bases[
                hint_bases_index_past_last_prev:hint_bases_index_past_last] = (
                hint_base_bases_direct)
        # Else, this pseudo-superclass is *NOT* a PEP 560-compliant generic.

        #FIXME: *NICE*. For safety, we should also be doing something similar in
        #our make_check_expr()-driven dynamic code generator as well.
        # Nullify the previously visited pseudo-superclass in this list,
        # avoiding spurious memory leaks by encouraging garbage collection.
        hint_bases[hint_bases_index_curr] = None

        # Increment the 0-based index of the next visited pseudo-superclass in
        # this list *BEFORE* visiting that pseudo-superclass but *AFTER*
        # performing all other logic for the current pseudo-superclass.
        hint_bases_index_curr += 1

    # ....................{ POSTAMBLE                      }....................
    # Release this list. Pray for salvation, for we find none here.
    release_fixed_list(hint_bases)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **PEP-compliant type hint sign getter** unit tests.

This submodule unit tests the public API of the private
:mod:`beartype._util.hint.pep.utilpepsign` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS                              }....................
def test_get_hint_pep_sign(hints_pep_meta) -> None:
    '''
    Test the
    :func:`beartype._util.hint.pep.utilpepget.get_hint_pep_sign` getter.

    Parameters
    ----------
    hints_pep_meta : List[beartype_test.a00_unit.data.hint.util.data_hintmetacls.HintPepMetadata]
        List of PEP-compliant type hint metadata describing sample PEP-compliant
        type hints exercising edge cases in the :mod:`beartype` codebase.
    '''

    # Defer test-specific imports.
    from beartype.roar import BeartypeDecorHintPepSignException
    from beartype._util.hint.pep.utilpepsign import get_hint_pep_sign
    from beartype_test.a00_unit.data.hint.data_hint import (
        HINTS_NONPEP,
        NonpepCustomFakeTyping,
    )
    from pytest import raises

    # Assert this getter returns the expected unsubscripted "typing" attribute
    # for all PEP-compliant type hints associated with such an attribute.
    for hint_pep_meta in hints_pep_meta:
        # Localize this hint to simplify debugging.
        hint = hint_pep_meta.hint
        assert get_hint_pep_sign(hint) is hint_pep_meta.pep_sign

    # Assert this getter raises the expected exception for an instance of a
    # class erroneously masquerading as a "typing" class.
    with raises(BeartypeDecorHintPepSignException):
        # Localize this return value to simplify debugging.
        hint_nonpep_sign = get_hint_pep_sign(NonpepCustomFakeTyping())

    # Assert this getter raises the expected exception for non-"typing" hints.
    for hint_nonpep in HINTS_NONPEP:
        with raises(BeartypeDecorHintPepSignException):
            # Localize this return value to simplify debugging.
            hint_nonpep_sign = get_hint_pep_sign(hint_nonpep)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Test-wide **main codebase paths** (i.e., low-level callables creating and
returning :class:`pathlib.Path` objects encapsulating test-agnostic paths
applicable to the codebase being tested).
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from beartype.meta import (
    PACKAGE_NAME,
    PACKAGE_TEST_NAME,
)
from beartype._util.cache.utilcachecall import callable_cached
from beartype_test._util.path.pytpathlib import (
    DirRelative,
    FileRelative,
)
from pathlib import Path

# ....................{ GETTERS ~ dir                      }....................
@callable_cached
def get_main_dir() -> Path:
    '''
    :class:`.Path` encapsulating the absolute dirname of the **top-level project
    directory** (i.e., directory containing both a ``.git/`` subdirectory and a
    subdirectory providing this project's package) if found *or* raise an
    exception otherwise.
    '''
    # print(f'current module paths: {__package__} [{__file__}]')

    # Path encapsulating the current module.
    MODULE_FILE = Path(__file__)

    # Path encapsulating the current module's package.
    MODULE_PACKAGE_DIR = MODULE_FILE.parent

    # Path encapsulating the relative dirname of this project's directory
    # relative to the dirname of the package defining the current module.
    MAIN_DIR = DirRelative(MODULE_PACKAGE_DIR, '../../..')

    # If this project's directory either does not contain a "beartype_test"
    # subdirectory *OR* does but this path is not a directory, raise an
    # exception. This basic sanity check improves the likelihood that this
    # project directory is what we assume it is.
    #
    # Note that we intentionally avoid testing paths *NOT* bundled with release
    # tarballs (e.g., a root ".git/" directory), as doing so would prevent
    # external users and tooling from running tests from release tarballs.
    DirRelative(MAIN_DIR, PACKAGE_TEST_NAME)

    # Return this path.
    return MAIN_DIR


@callable_cached
def get_main_package_dir() -> Path:
    '''
    :mod:`Path` encapsulating the absolute dirname of the **top-level project
    package** (i.e., directory providing this package's top-level package
    containing at least an ``__init__.py`` file) if found *or* raise an
    exception otherwise.
    '''

    # Terrifying terseness!
    return DirRelative(get_main_dir(), PACKAGE_NAME)

# ....................{ GETTERS ~ dir : sphinx             }....................
@callable_cached
def get_main_sphinx_dir() -> Path:
    '''
    :mod:`Path` encapsulating the absolute dirname of the **top-level Sphinx
    documentation tree** (i.e., directory containing both the input and output
    subdirectories leveraged by Sphinx to build documentation for this project)
    if found *or* raise an exception otherwise.
    '''

    # Immense propensity!
    return DirRelative(get_main_dir(), 'doc')


@callable_cached
def get_main_sphinx_source_dir() -> Path:
    '''
    :mod:`Path` encapsulating the absolute dirname of the **source Sphinx
    documentation tree** (i.e., directory containing the input subdirectory
    leveraged by Sphinx to build documentation for this project) if found *or*
    raise an exception otherwise.
    '''

    # Immense propensity!
    return DirRelative(get_main_sphinx_dir(), 'src')

# ....................{ GETTERS ~ file                     }....................
@callable_cached
def get_main_mypy_config_file() -> Path:
    '''
    :mod:`Path` encapsulating the absolute filename of this project's **mypy
    configuration file** (i.e., top-level ``.mypy.ini`` file) if found *or*
    raise an exception otherwise.
    '''

    # Obverse obviation!
    return FileRelative(get_main_dir(), 'mypy.ini')


@callable_cached
def get_main_readme_file() -> Path:
    '''
    :mod:`Path` encapsulating the absolute filename of the **project readme
    file** (i.e., this project's front-facing ``README.rst`` file) if found *or*
    raise an exception otherwise.

    Note that the :meth:`.Path.read_text` method of this object trivially yields
    the decoded plaintext contents of this file as a string.
    '''

    # Perverse pomposity!
    return FileRelative(get_main_dir(), 'README.rst')



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **chain set class hierarchy** (i.e., private classes implementing
mutable set-like chain mappings dynamically compositing the keys of two or more
mutable dictionaries treated as sets).
'''

# ....................{ IMPORTS                            }....................
from collections import ChainMap

# ....................{ CLASSES                            }....................
class ChainSet(ChainMap):
    '''
    **Chain set** (i.e., mutable set-like chain mappings dynamically compositing
    the keys of two or more mutable dictionaries treated as sets).

    This data structure is currently simply a trivial subclass of the standard
    :class:`.ChainMap` type. Although non-ideal, this subclass enables callers
    to at least differentiate between standard chain maps and set-like chain
    maps. Ideally, this subclass would be entirely reimplemented "from the
    ground up" *without* reference to the semantically unrelated
    :class:`.ChainMap` type. Until then, this approach perseveres.
    '''

    pass



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **NumPy test data fixtures** (i.e., high-level session-scoped
:mod:`pytest` fixtures exposing various third-party :mod:`numpy` objects,
exercising edge cases in unit tests requiring these fixtures).
'''

# ....................{ IMPORTS                            }....................
from pytest import fixture

# ....................{ CLASSES                            }....................
class _NumpyArrays(object):
    '''
    **NumPy arrays dataclass** (i.e., object yielded by the
    :func:`.numpy_arrays` fixture comprising various third-party :mod:`numpy`
    arrays of interest to downstream unit tests).

    Attributes
    ----------
    arrays : tuple[numpy.ndarray]
        Tuple of all NumPy arrays classified by this dataclass.
    array_1d_boolean : numpy.typing.NDArray[numpy.bool_]
        One-dimensional NumPy array containing only booleans.
    array_1d_complex_128 : numpy.typing.NDArray[numpy.complex128]
        One-dimensional NumPy array containing only 128-bit complex numbers.
    array_1d_float_32 : numpy.typing.NDArray[numpy.float32]
        One-dimensional NumPy array containing only 32-bit floating-point
        numbers.
    array_1d_float_64 : numpy.typing.NDArray[numpy.float64]
        One-dimensional NumPy array containing only 64-bit floating-point
        numbers.
    array_1d_int_32 : numpy.typing.NDArray[numpy.int32]
        One-dimensional NumPy array containing only 32-bit signed integers.
    array_1d_int_64 : numpy.typing.NDArray[numpy.int64]
        One-dimensional NumPy array containing only 64-bit signed integers.
    array_1d_uint_32 : numpy.typing.NDArray[numpy.uint32]
        One-dimensional NumPy array containing only 32-bit unsigned
        integers.
    array_1d_uint_64 : numpy.typing.NDArray[numpy.uint64]
        One-dimensional NumPy array containing only 64-bit unsigned
        integers.
    array_1d_memory_view : numpy.typing.NDArray[numpy.void]
        One-dimensional NumPy array serving as a memory view over arbitrary
        bytes all of the same length.
    array_1d_string_byte : numpy.typing.NDArray[numpy.str_]
        One-dimensional NumPy array containing only byte strings.
    array_1d_string_char : numpy.typing.NDArray[numpy.bytes_]
        One-dimensional NumPy array containing only Unicode strings.
    '''

    # ....................{ INITIALIZERS                   }....................
    def __init__(self) -> None:
        '''
        Initialize this NumPy arrays dataclass.
        '''

        # Defer fixture-specific imports.
        from numpy import (
            asarray,
            complex128,
            float32,
            float64,
            int32,
            int64,
            uint32,
            uint64,
            void,
        )

        # NumPy array containing only booleans.
        self.array_1d_boolean = asarray(
            (True, False, True, True, False, True, False, False,))

        # NumPy array containing only 128-bit complex numbers.
        self.array_1d_complex_128 = asarray(
            (1.3 + 8.23j, 70.222 + 726.2431j, 8294.28730 + 100776.357238j),
            dtype=complex128,
        )

        # NumPy array containing only 32-bit floats.
        self.array_1d_float_32 = asarray(
            (1.2, 2.4, 3.0, 3.6, 4.0, 4.5, 4.8, 5.6, 6.0, 6.3, 7.0,),
            dtype=float32,
        )

        # NumPy array containing only 64-bit floats.
        self.array_1d_float_64 = asarray(
            (3.2, 5, 1, 2, 1, 8, 2, 5, 1, 3, 1, 2.8, 1, 1.5, 1, 1, 4,),
            dtype=float64,
        )

        # NumPy array containing only 32-bit signed integers.
        self.array_1d_int_32 = asarray(
            (1, 0, -3, -5, 2, 6, -4, 9, -2, 3, 8, -4, 1, -3, -7, -7, 5, 0,),
            dtype=int32,
        )

        # NumPy array containing only 64-bit signed integers.
        self.array_1d_int_64 = asarray(
            (1, 7, -39, 211, 1168, -6728, -40561, -256297, 1696707,),
            dtype=int64,
        )

        # NumPy array containing only 32-bit unsigned integers.
        self.array_1d_uint_32 = asarray(
            (47, 49, 46, 38, 39, 61),
            dtype=uint32,
        )

        # NumPy array containing only 64-bit unsigned uintegers.
        self.array_1d_uint_64 = asarray(
            (4, 36, 624, 3744, 5108, 10200, 54912,),
            dtype=uint64,
        )

        # NumPy array serving as a memory view over arbitrary bytes all of the
        # same length.
        self.array_1d_memory_view = asarray(
            (
                b'The fissured stones with its entwining arms,',
                b'And did embower with leaves for ever green, ',
                b'And berries dark, the smooth and even space ',
                b'Of its inviolated floor, and here the childr',
            ),
            dtype=void,
        )

        # NumPy array containing only byte strings.
        self.array_1d_string_byte = asarray((
            b'It overlooked in its serenity',
            b'The dark earth, and the bending vault of stars.',
            b'It was a tranquil spot, that seemed to smile',
            b'Even in the lap of horror. Ivy clasped',
        ))

        # NumPy array containing only Unicode strings.
        self.array_1d_string_char = asarray((
            'Yet the grey precipice and solemn pine',
            'And torrent, were not all;—one silent nook',
            'Was there. Even on the edge of that vast mountain,',
            'Upheld by knotty roots and fallen rocks,',
        ))

        # Tuple of *ALL* of the above NumPy arrays.
        self.arrays = (
            self.array_1d_boolean,
            self.array_1d_complex_128,
            self.array_1d_float_32,
            self.array_1d_float_64,
            self.array_1d_int_32,
            self.array_1d_int_64,
            self.array_1d_uint_32,
            self.array_1d_uint_64,
            self.array_1d_memory_view,
            self.array_1d_string_byte,
            self.array_1d_string_char,
        )

# ....................{ FIXTURES                           }....................
@fixture(scope='session')
def numpy_arrays() -> _NumpyArrays:
    '''
    Session-scoped fixture yielding a **NumPy arrays dataclass** (i.e.,
    object comprising various third-party :mod:`numpy` arrays of interest to
    higher-level unit tests).

    Raises
    ------
    :exc:`pytest.Skipped`
        If :mod:`numpy` is unimportable under the active Python interpreter, in
        which case *all* tests transitively requiring this fixture will be
        skipped as well. Praise be to :mod:`pytest`.

    See Also
    --------
    https://stackoverflow.com/a/46854903/2809027
        StackOverflow answer lightly inspiring this implementation.
    '''

    # ..................{ PREAMBLE                           }..................
    # Defer preamble-specific imports.
    from beartype_test._util.module.pytmodtest import is_package_numpy
    from pytest import skip

    # If NumPy is unimportable, raise the "pytest.Skipped" exception skipping
    # *ALL* tests transitively requiring this fixture.
    #
    # Note that explicitly calling "skip_unless_package('numpy')" here fails to
    # raise the expected exception, even when NumPy is unimportable. Why?
    # Because that call only behaves as expected in a decorator context. *sigh*
    if not is_package_numpy():
        skip('NumPy unimportable.')
    # Else, NumPy is importable.

    # ..................{ YIELD                              }..................
    # Yield a singleton instance of this NumPy array dataclass.
    yield _NumpyArrays()



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **Decidedly Object-Oriented Runtime-checking (DOOR) fixtures** (i.e.,
:mod:`pytest`-specific context managers passed as parameters to unit tests
exercising the :mod:`beartype.door.is_subhint` function).
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from pytest import fixture

# ....................{ FIXTURES                           }....................
@fixture(scope='session')
def door_cases_is_subhint() -> 'Iterable[Tuple[object, object, bool]]':
    '''
    Session-scoped fixture returning an iterable of **type subhint cases**
    (i.e., 3-tuples ``(subhint, superhint, is_subhint)`` describing the subhint
    relations between two type hints), efficiently cached across all tests
    requiring this fixture.

    This iterable is intentionally defined by the return of this fixture rather
    than as a global constant of this submodule. Why? Because the former safely
    defers all heavyweight imports required to define this iterable to the call
    of the first unit test requiring this fixture, whereas the latter unsafely
    performs those imports at pytest test collection time.

    Returns
    -------
    Iterable[Tuple[object, object, bool]]
        Iterable of one or more 3-tuples ``(subhint, superhint, is_subhint)``,
        where:

        * ``subhint`` is the type hint to be passed as the first parameter to
          the :func:`beartype.door.is_subhint` tester.
        * ``superhint`` is the type hint to be passed as the second parameter to
          the :func:`beartype.door.is_subhint` tester.
        * ``is_subhint`` is :data:`True` only if that subhint is actually a
          subhint of that superhint according to that tester.
    '''

    # ..................{ IMPORTS                            }..................
    # Defer fixture-specific imports.
    import collections.abc
    import typing
    from abc import ABCMeta
    from beartype._util.cls.utilclstest import is_type_subclass_proper
    from beartype._util.py.utilpyversion import IS_PYTHON_AT_MOST_3_13
    from beartype_test.a00_unit.data.pep.generic.data_pep484generic import (
        Pep484GenericT,
        Pep484GenericSubT,
        Pep484GenericST,
        Pep484GenericSInt,
    )
    from beartype_test.a00_unit.data.pep.generic.data_pep585generic import (
        Pep585SequenceT,
    )
    from beartype_test.a00_unit.data.pep.data_pep484 import (
        S,
        T,
        T_sequence,
        T_int_or_str,
    )
    from collections.abc import (
        Collection as CollectionABC,
        Sequence as SequenceABC,
    )

    # Intentionally import from "beartype.typing" rather than "typing" to
    # guarantee PEP 544-compliant caching protocol type hints.
    from beartype.typing import (
        Annotated,
        Any,
        Awaitable,
        Callable,
        Collection,
        DefaultDict,
        Dict,
        Hashable,
        Iterable,
        List,
        Literal,
        Mapping,
        NamedTuple,
        NewType,
        Optional,
        Protocol,
        Reversible,
        Sequence,
        Sized,
        Tuple,
        Type,
        TypedDict,
        Union,
    )

    # ..................{ NEWTYPES                           }..................
    NewStr = NewType('NewStr', str)

    # ..................{ CLASSES                            }..................
    class MuhThing:
        def muh_method(self):
            pass

    class MuhSubThing(MuhThing):
        pass

    class MuhNutherThing:
        def __len__(self) -> int:
            pass


    class MuhDict(TypedDict):
        '''
        Arbitrary typed dictionary.
        '''

        thing_one: str
        thing_two: int


    class MuhThingP(Protocol):
        '''
        Arbitrary caching @beartype protocol.
        '''

        def muh_method(self):
            ...


    class MuhTuple(NamedTuple):
        '''
        Arbitrary named tuple.
        '''

        thing_one: str
        thing_two: int

    # ..................{ LISTS ~ cases                      }..................
    # List of all hint subhint cases (i.e., 3-tuples "(subhint, superhint,
    # is_subhint)" describing the subhint relations between two PEP-compliant
    # type hints) to be returned by this fixture.
    HINT_SUBHINT_CASES = [
        # ..................{ PEP 484 ~ argless : any        }..................
        # PEP 484-compliant catch-all type hint.
        (MuhThing, Any, True),
        (Tuple[object, ...], Any, True),
        (Union[int, MuhThing], Any, True),

        # Although *ALL* type hints are subhints of "Any", "Any" is *ONLY* a
        # subhint of itself.
        (Any, Any, True),
        (Any, object, False),

        # ..................{ PEP 484 ~ argless : bare       }..................
        # PEP 484-compliant unsubscripted type hints, which are necessarily
        # subhints of themselves.
        (list, list, True),
        (list, List, True),

        # PEP 484-compliant unsubscripted sequence type hints.
        (Sequence, List, False),
        (Sequence, list, False),
        (List, Sequence, True),
        (list, Sequence, True),
        (list, SequenceABC, True),
        (list, CollectionABC, True),

        # ..................{ PEP 484 ~ argless : type       }..................
        # PEP 484-compliant argumentless abstract base classes (ABCs).
        (str, Hashable, True),
        (MuhNutherThing, Sized, True),
        (MuhTuple, tuple, True),  # not really types

        # PEP 484-compliant new type type hints.
        (NewStr, NewStr, True),
        (NewStr, int, False),
        (NewStr, str, True),
        (int, NewStr, False),
        (str, NewStr, False),  # NewType act like subtypes

        # ..................{ PEP 484 ~ argless : typevar    }..................
        # PEP 484-compliant type variables.
        (list, T_sequence, True),
        (T_sequence, list, False),
        (int, T_int_or_str, True),
        (str, T_int_or_str, True),
        (list, T_int_or_str, False),
        (Union[int, str], T_int_or_str, True),
        (Union[int, str, None], T_int_or_str, False),
        (T, T_sequence, False),
        (T_sequence, T, True),
        (T_sequence, Any, True),
        (Any, T, True),  # Any is compatible with an unconstrained TypeVar
        (Any, T_sequence, False),  # but not vice versa

        # ..................{ PEP 484 ~ argless : number     }..................
        # Blame Guido.
        (bool, int, True),

        # PEP 484-compliant implicit numeric tower, which we explicitly and
        # intentionally do *NOT* comply with. Floats are not integers. Notably,
        # floats *CANNOT* losslessly represent many integers and are thus
        # incompatible in general.
        (float, int, False),
        (complex, int, False),
        (complex, float, False),
        (int, float, False),
        (float, complex, False),

        # ..................{ PEP 484 ~ generic              }..................
        # "typing.Generic"-centric tests.

        #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        # CAUTION: Synchronize changes below by replacing all references in
        # the first tuple item to "Pep484GenericT" with "Pep484GenericSubT".
        #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        # PEP 484-compliant generic superclasses parametrized by one
        # unconstrained type variables.
        (Pep484GenericT, Pep484GenericT, True),
        (Pep484GenericT, Pep484GenericT[int], False),
        (Pep484GenericT[int], Pep484GenericT, True),
        (Pep484GenericT[int], Pep484GenericT[T_sequence], False),
        (Pep484GenericT[list], Pep484GenericT[T_sequence], True),
        (Pep484GenericT[list], Pep484GenericT[Sequence], True),
        (Pep484GenericT[str], Pep484GenericT[T_sequence], True),
        (Pep484GenericT[Sequence], Pep484GenericT[list], False),
        (Pep484GenericT[T_sequence], Pep484GenericT, True),

        #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        # CAUTION: Synchronize changes above.
        #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        # PEP 484-compliant generic subclasses parametrized by one unconstrained
        # type variables.
        (Pep484GenericSubT, Pep484GenericT, True),
        (Pep484GenericSubT, Pep484GenericT[int], False),
        (Pep484GenericSubT[int], Pep484GenericT, True),
        (Pep484GenericSubT[int], Pep484GenericT[T_sequence], False),
        (Pep484GenericSubT[list], Pep484GenericT[T_sequence], True),
        (Pep484GenericSubT[list], Pep484GenericT[Sequence], True),
        (Pep484GenericSubT[str], Pep484GenericT[T_sequence], True),
        (Pep484GenericSubT[Sequence], Pep484GenericT[list], False),
        (Pep484GenericSubT[T_sequence], Pep484GenericT, True),

        # PEP 484-compliant generic subclasses parametrized by one unconstrained
        # type variables and one concrete type.
        (Pep484GenericSInt, Pep484GenericST, True),
        (Pep484GenericSInt, Pep484GenericST[int, int], False),
        (Pep484GenericSInt[int], Pep484GenericST, True),
        (Pep484GenericSInt[int], Pep484GenericST[S, T_sequence], False),
        (Pep484GenericSInt[list], Pep484GenericST[T_sequence, object], True),
        (Pep484GenericSInt[list], Pep484GenericST[Sequence, Any], True),
        (Pep484GenericSInt[str], Pep484GenericST[T_sequence, S], True),
        (Pep484GenericSInt[Sequence], Pep484GenericST[list, object], False),
        (Pep484GenericSInt[T_sequence], Pep484GenericST, True),

        # ..................{ PEP 484 ~ optional             }..................
        # "typing.Optional"-centric tests.

        # PEP 484-compliant optionals.
        (int, Optional[int], True),
        (Optional[int], int, False),
        (list, Optional[Sequence], True),

        # ..................{ PEP 484 ~ union                }..................
        # "typing.Union"-centric tests.

        # PEP 484-compliant unions.
        (int, Union[int, str], True),
        (Union[int, str], Union[list, int, str], True),
        (Union[str, int], Union[int, str, list], True),  # order doesn't matter
        (Union[str, list], Union[str, int], False),
        (Union[int, str, list], list, False),
        (Union[int, str, list], Union[int, str], False),

        # ..................{ PEP (484|585) ~ callable       }..................
        # PEP 484-compliant callable type hints.
        (Callable, Callable[..., Any], True),
        (Callable[[], int], Callable[..., Any], True),
        (Callable[[int, str], List[int]], Callable, True),
        (Callable[[int, str], List[int]], Callable, True),
        (
            Callable[[float, List[str]], int],
            Callable[[int, Sequence[str]], int],
            True,
        ),
        (Callable[[Sequence], int], Callable[[list], int], False),
        (Callable[[], int], Callable[..., None], False),
        (Callable[..., Any], Callable[..., None], False),
        (Callable[[float], None], Callable[[float, int], None], False),
        (Callable[[], int], Sequence[int], False),
        (Callable[[int, str], int], Callable[[int, str], Any], True),
        # (types.FunctionType, Callable, True),  # FIXME

        # ..................{ PEP (484|585) ~ generic        }..................
        # PEP 484- or 585-compliant generic subclasses inheriting PEP 484- or
        # 585-compliant type hints parametrized by one unconstrained type
        # variables.
        (Pep585SequenceT, Sequence[T], True),
        (Pep585SequenceT, Sequence[int], False),
        (Pep585SequenceT[int], Sequence[T], True),
        (Pep585SequenceT[int], Sequence[T_sequence], False),
        (Pep585SequenceT[list], Sequence[T_sequence], True),
        (Pep585SequenceT[list], Sequence[Sequence], True),
        (Pep585SequenceT[str], Sequence[T_sequence], True),
        (Pep585SequenceT[Sequence], Sequence[list], False),
        (Pep585SequenceT[T_sequence], Sequence[T], True),

        # ..................{ PEP (484|585) ~ mapping        }..................
        # PEP 484-compliant mapping type hints.
        (dict, Dict, True),
        (Dict[str, int], Dict, True),
        (dict, Dict[str, int], False),
        (
            DefaultDict[str, Sequence[int]],
            Mapping[Union[str, int], Iterable[Union[int, str]]],
            True,
        ),

        # ..................{ PEP (484|585) ~ sequence       }..................
        # PEP 484-compliant sequence type hints.
        (List[int], List[int], True),
        (List[int], Sequence[int], True),
        (Sequence[int], Iterable[int], True),
        (Iterable[int], Sequence[int], False),
        (Sequence[int], Reversible[int], True),
        (Sequence[int], Reversible[str], False),
        (Collection[int], Sized, True),
        (List[int], List, True),  # if the super is un-subscripted, assume Any
        (List[int], List[Any], True),
        (Awaitable, Awaitable[str], False),
        (List[int], List[str], False),

        # PEP 484-compliant tuple type hints.
        (tuple, Tuple, True),
        (Tuple, Tuple, True),
        (tuple, Tuple[Any, ...], True),
        (tuple, Tuple[()], False),
        (Tuple[()], tuple, True),
        (Tuple[int, str], Tuple[int, str], True),
        (Tuple[int, str], Tuple[int, str, int], False),
        (Tuple[int, str], Tuple[int, Union[int, list]], False),
        (Tuple[Union[int, str], ...], Tuple[int, str], False),
        (Tuple[int, str], Tuple[str, ...], False),
        (Tuple[int, str], Tuple[Union[int, str], ...], True),
        (Tuple[Union[int, str], ...], Tuple[Union[int, str], ...], True),
        (Tuple[int], Dict[str, int], False),
        (Tuple[Any, ...], Tuple[str, int], False),

        # PEP 484-compliant nested sequence type hints.
        (List[int], Union[str, List[Union[int, str]]], True),

        # ..................{ PEP (484|585) ~ subclass       }..................
        # PEP 484-compliant subclass type hints.
        (Type[int], Type[int], True),
        (Type[int], Type[str], False),
        (Type[MuhSubThing], Type[MuhThing], True),
        (Type[MuhThing], Type[MuhSubThing], False),
        (MuhThing, Type[MuhThing], False),

        # ..................{ PEP 544                        }..................
        # PEP 544-compliant type hints.
        (MuhThing, MuhThingP, True),
        (MuhNutherThing, MuhThingP, False),
        (MuhThingP, MuhThing, False),

        # ..................{ PEP 585                        }..................
        # PEP 585-compliant type hints.
        (tuple, Tuple, True),
        (tuple[()], Tuple[()], True),

        # ..................{ PEP 586                        }..................
        # PEP 586-compliant type hints.
        (Literal[7], int, True),
        (Literal["a"], str, True),
        (Literal[7, 8, "3"], Union[int, str], True),
        (Literal[7, 8, "3"], Union[list, int], False),
        (Literal[True], Union[Literal[True], Literal[False]], True),
        (Literal[7, 8], Literal[7, 8, 9], True),
        (int, Literal[7], False),
        (Union[Literal[True], Literal[False]], Literal[True], False),

        # ..................{ PEP 589                        }..................
        # PEP 589-compliant type hints.
        (MuhDict, dict, True),

        # ..................{ PEP 593                        }..................
        # PEP 593-compliant type hints.
        (Annotated[int, "a note"], int, True),  # annotated is subtype of unannotated
        (int, Annotated[int, "a note"], False),  # but not vice versa
        (Annotated[list, True], Annotated[Sequence, True], True),
        (Annotated[list, False], Annotated[Sequence, True], False),
        (Annotated[list, 0, 0], Annotated[list, 0], False),  # must have same num args
        (Annotated[List[int], "metadata"], List[int], True),
    ]

    # ..................{ LISTS ~ cases : version            }..................
    # If the active Python interpreter targets Python <= 3.13...
    if IS_PYTHON_AT_MOST_3_13:
        # "collections.abc.ByteString", "typing.ByteString", and therefore
        # "beartype.typing.ByteString" itself no longer exist under Python 3.14.
        from beartype.typing import ByteString

        # Extend this list with...
        HINT_SUBHINT_CASES.extend((
            # ..................{ PEP 484 ~ argless : type   }..................
            # PEP 484-compliant argumentless abstract base classes (ABCs).
            (bytes, ByteString, True),
        ))

    # ..................{ LISTS ~ typing                     }..................
    # List of the unqualified basenames of all standard ABCs published by
    # the standard "collections.abc" module, initialized to the empty list.
    COLLECTIONS_ABC_BASENAMES = []

    #FIXME: Call the higher-level and *SIGNIFICANTLY* safer
    #get_object_attrs_name_to_value_explicit() getter instead of manually
    #iterating over "__dict__.items()", please. *sigh*
    # For the unqualified basename of each attribute and that attribute defined
    # by the standard "collections.abc" submodule...
    for collections_abc_basename, collections_abc in (
        collections.abc.__dict__.items()):
        # If this attribute is a public ABC, include this ABC, where public ABCs
        # are detected as...
        if (
            # The unqualified basename of this attribute is not prefixed by an
            # underscore (and is thus public) *AND*...
            not collections_abc_basename.startswith('_') and
            # This attribute is an ABC but *NOT* the semantically meaningless
            # "abc.ABCMeta" superclass itself.
            is_type_subclass_proper(collections_abc, ABCMeta)
        ):
            COLLECTIONS_ABC_BASENAMES.append(collections_abc_basename)
        # Else, this is an unrelated attribute. In this case, silently ignore
        # this attribute and continue to the next.

    # List of the unqualified basenames of all standard abstract base classes
    # (ABCs) supported by the standard "typing" module, defined as the
    # concatenation of...
    TYPING_ABC_BASENAMES = (
        # List of the unqualified basenames of all standard ABCs published by
        # the standard "collections.abc" module *PLUS*...
        COLLECTIONS_ABC_BASENAMES +
        # List of the unqualified basenames of all ancillary ABCs *NOT*
        # published by the standard "collections.abc" module but nonetheless
        # supported by the standard "typing" module.
        ['Deque',]
    )
    # print(f'TYPING_ABC_BASENAMES: {TYPING_ABC_BASENAMES}')

    # ..................{ HINTS ~ abcs                       }..................
    # For the unqualified basename of each standard ABCs supported by the
    # standard "typing" module...
    #
    # Note this also constitutes a smoke test (i.e., high-level test validating
    # core functionality) for whether the DOOR API supports standard abstract
    # base classes (ABCs). Smoke out those API inconsistencies, pytest!
    for TYPING_ABC_BASENAME in TYPING_ABC_BASENAMES:
        #FIXME: This logic is likely to fail under a future Python release.
        # Type hint factory published by the "typing" module corresponding to
        # this ABC if any *OR* "None" otherwise (i.e., if "typing" publishes
        # *NO* such type hint factory).
        typing_abc = getattr(typing, TYPING_ABC_BASENAME, None)

        # If "typing" publishes *NO* such type hint factory, silently ignore
        # this ABC and continue to the next.
        if typing_abc is None:
            continue
        # Else, "typing" publishes this type hint factory.

        # Number of type variables parametrizing this ABC, defined as a private
        # instance variable of this type hint factory yielding this metadata.
        # Under Python >= 3.9, unsubscripted type hint factories are *NOT*
        # parametrized by type variables.
        typing_abc_typevars_len = typing_abc._nparams

        # If this ABC is parametrized by one or more type variables, exercise
        # that this ABC subscripted by one or more arbitrary concrete types is a
        # non-trivial subhint of this same ABC subscripted by one or more
        # arbitrary different ABCs of those concrete types.
        if typing_abc_typevars_len:
            subhint =   typing_abc[(list,)     * typing_abc_typevars_len]
            superhint = typing_abc[(Sequence,) * typing_abc_typevars_len]
        # Else, this ABC is parametrized by *NO* type variables. In this case,
        # fallback to exercising that this ABC is a trivial subhint of itself.
        else:
            subhint =   typing_abc
            superhint = typing_abc

        # Append a new hint subhint case exercising that this subhint is
        # actually a subhint of this superhint.
        HINT_SUBHINT_CASES.append((subhint, superhint, True))

    # ..................{ RETURN                             }..................
    # Return this mutable list coerced into an immutable tuple for safety.
    return tuple(HINT_SUBHINT_CASES)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype decorator :mod:`contextlib`-specific unit tests.

This submodule unit tests the :func:`beartype.beartype` decorator with respect
the standard :mod:`contextlib` module.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from beartype_test._util.mark.pytskip import skip_if_python_version_less_than

# ....................{ TESTS                              }....................
@skip_if_python_version_less_than('3.11.0')
def test_decor_contextlib_contextmanager() -> None:
    '''
    Test the :func:`beartype.beartype` decorator on
    :func:`contextlib.contextmanager`-based **synchronous context managers**
    (i.e., generator factory functions decorated by that standard decorator) if
    the active Python interpreter targets Python >= 3.11 and thus defines the
    ``co_qualname`` attribute on code objects required to implement this
    functionality *or* skip this test otherwise.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype import beartype
    from beartype.roar import BeartypeCallHintParamViolation
    from beartype.typing import (
        Iterator,
        Union,
    )
    from contextlib import contextmanager
    from pytest import raises

    # ....................{ CONTEXTS                       }....................
    @contextmanager
    @beartype
    def and_motions_of(
        the_forests_and_the_sea: Union[int, float]) -> Iterator[int]:
        '''
        Arbitrary :func:`contextlib.contextmanager`-decorated context manager
        decorated by :func:`beartype.beartype` in the ideal order.
        '''

        yield int(the_forests_and_the_sea)


    @beartype
    @contextmanager
    def may_modulate_with(
        murmurs_of_the_air: str) -> Iterator[Union[str, bool]]:
        '''
        Arbitrary :func:`contextlib.contextmanager`-decorated context manager
        decorated by :func:`beartype.beartype` in a non-ideal order.
        '''

        yield murmurs_of_the_air

    # ....................{ PASS                           }....................
    # Assert that the ideal context manager when passed a valid parameter
    # returns the expected return.
    with and_motions_of(len(
        'Of night and day, and the deep heart of man.')) as (
        deep_heart_of_man):
        assert deep_heart_of_man == len(
            'Of night and day, and the deep heart of man.')

    # Assert that the non-ideal context manager when passed a valid parameter
    # returns the expected return.
    with may_modulate_with(
        'And voice of living beings, and woven hymns') as and_woven_hymns:
        assert and_woven_hymns == 'And voice of living beings, and woven hymns'

    # ....................{ FAIL                           }....................
    # Assert that the ideal context manager when passed an invalid parameter
    # raises the expected exception.
    with raises(BeartypeCallHintParamViolation):
        with and_motions_of('No human hands with pious reverence reared,'):
            pass

    # Assert that the non-ideal context manager when passed an invalid parameter
    # raises the expected exception.
    with raises(BeartypeCallHintParamViolation):
        with may_modulate_with(b'There was a Poet whose untimely tomb'):
            pass


@skip_if_python_version_less_than('3.11.0')
async def test_decor_contextlib_asynccontextmanager() -> None:
    '''
    Test the :func:`beartype.beartype` decorator on
    :func:`contextlib.asynccontextmanager`-based **asynchronous context
    managers** (i.e., generator factory functions decorated by that standard
    decorator) if the active Python interpreter targets Python >= 3.11 and thus
    defines the ``co_qualname`` attribute on code objects required to implement
    this functionality *or* skip this test otherwise.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype import beartype
    from beartype.roar import BeartypeCallHintParamViolation
    from beartype.typing import (
        AsyncIterator,
        Union,
    )
    from contextlib import asynccontextmanager
    from pytest import raises

    # ....................{ CONTEXTS                       }....................
    @asynccontextmanager
    @beartype
    async def art_king_of(
        this_frail_world: Union[int, float]) -> AsyncIterator[int]:
        '''
        Arbitrary :func:`contextlib.asynccontextmanager`-decorated context
        manager decorated by :func:`beartype.beartype` in the ideal order.
        '''

        yield int(this_frail_world)


    @beartype
    @asynccontextmanager
    async def from_the_red_field(
        of_slaughter: str) -> AsyncIterator[Union[str, bool]]:
        '''
        Arbitrary :func:`contextlib.asynccontextmanager`-decorated context
        manager decorated by :func:`beartype.beartype` in a non-ideal order.
        '''

        yield of_slaughter

    # ....................{ PASS                           }....................
    # Assert that the ideal context manager when passed a valid parameter
    # returns the expected return.
    async with art_king_of(len(
        'Art king of this frail world, from the red field')) as (
        from_the_reeking_hospital):
        assert from_the_reeking_hospital == len(
            'Art king of this frail world, from the red field')

    # Assert that the non-ideal context manager when passed a valid parameter
    # returns the expected return.
    async with from_the_red_field(
        'Of slaughter, from the reeking hospital,') as (
        the_patriots_sacred_couch):
        assert the_patriots_sacred_couch == (
            'Of slaughter, from the reeking hospital,')

    # ....................{ FAIL                           }....................
    # Assert that the ideal context manager when passed an invalid parameter
    # raises the expected exception.
    with raises(BeartypeCallHintParamViolation):
        async with art_king_of("The patriot's sacred couch, the snowy bed"):
            pass

    # Assert that the non-ideal context manager when passed an invalid parameter
    # raises the expected exception.
    with raises(BeartypeCallHintParamViolation):
        async with from_the_red_field(
            b'Of innocence, the scaffold and the throne,'):
            pass
